# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

import contextlib

import numpy as np

from PySide6 import QtCore
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QSlider,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from . import configbase


cb = configbase  # Workaround for Python 3.6 issue 30024


def _limits_for_qt(set_limits):
    limits = [-1e9, 1e9]

    if set_limits is not None:
        for i in range(2):
            if set_limits[i] is not None:
                limits[i] = set_limits[i]

    return limits


def try_rst_to_html(s):
    try:
        return rst_to_html(s)
    except Exception:
        return s


def rst_to_html(s):
    import re

    from docutils.core import publish_parts

    s = re.sub(r":ref:`([\s\S]+?)\s*<([\s\S]+)>`", r"\1", s, re.MULTILINE)
    parts = publish_parts(s, writer_name="html")
    return parts["body"]


def wrap_qwidget(cls):
    class QWidgetWrapper(cls):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

        def wheelEvent(self, event):
            if self.hasFocus():
                super().wheelEvent(event)
            else:
                event.ignore()

    return QWidgetWrapper


class Pidget(QFrame):
    def __init__(self, param, parent_instance):
        super().__init__()

        self.param = param
        self._parent_instance = parent_instance

        self._enable_event_handler = False

        self.hide()

    def update(self, alerts=None):
        self._enable_event_handler = False
        self._update(alerts)
        self._enable_event_handler = True

    def _update(self, *args, **kwargs):
        pass

    def _get_param_value(self):
        return self.param.__get__(self._parent_instance)

    def _subwidget_event_handler(self, val):
        if self._enable_event_handler:
            self.param.pidget_event_handler(self._parent_instance, val)


class PidgetStub(Pidget):
    def __init__(self, param, parent_instance):
        super().__init__(param, parent_instance)

        self.setObjectName("frame")
        self.default_css = "#frame {border: 1px solid lightgrey; border-radius: 3px;}"
        self.setStyleSheet(self.default_css)

        doc = param.generate_doc()
        if doc:
            doc = try_rst_to_html(doc)
            self.setToolTip(doc)

        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 6, 0, 6)

        self.grid_widget = QWidget()
        self.layout.addWidget(self.grid_widget)

        self.grid = QGridLayout(self.grid_widget)
        self.grid.setContentsMargins(6, 0, 6, 0)

        self.alert_frame = QFrame()
        self.alert_frame.setStyleSheet(".QFrame {border-top: 1px solid lightgrey;}")
        alert_layout = QHBoxLayout(self.alert_frame)
        alert_layout.setContentsMargins(6, 6, 6, 0)
        self.alert_label = QLabel()
        self.alert_label.setStyleSheet("color: #333333; font-weight: 600;")
        alert_layout.addWidget(self.alert_label)
        self.layout.addWidget(self.alert_frame)
        self._set_alert(None)

    def _update(self, alerts):
        state = self._parent_instance._state

        if state != cb.Config.State.UNLOADED:
            if callable(self.param.visible):
                visible = self.param.visible(self._parent_instance)
            else:
                visible = self.param.visible
        else:
            visible = False

        self.setVisible(bool(visible))

        enabled = (
            state == cb.Config.State.LOADED
            or state == cb.Config.State.LIVE
            and self.param.is_live_updateable
        )
        enabled = enabled and self.param.enabled
        self.setEnabled(enabled)

        self._set_alert(alerts)

    def _set_alert(self, alerts):
        if not alerts:
            self.alert_frame.hide()
            self.setStyleSheet(self.default_css)
            return

        if isinstance(alerts, cb.Alert):
            alert = alerts
        else:
            alerts = sorted(alerts, key=lambda a: a.severity)
            alert = alerts[0]

        if alert.severity == cb.Severity.ERROR:
            bg = "FFB9A8"
        elif alert.severity == cb.Severity.WARNING:
            bg = "FFDFA8"
        else:
            bg = "FFFFEE"

        self.setStyleSheet(
            (
                "#frame {{"
                "background-color: #{};"
                "border: 1px solid lightgrey;"
                "border-radius: 3px;"
                "}}"
            ).format(bg)
        )
        self.alert_label.setText(alert.msg)
        self.alert_frame.show()

    def _subwidget_event_handler(self, val):
        with contextlib.suppress(ValueError):
            super()._subwidget_event_handler(val)


class ComboBoxPidget(PidgetStub):
    def __init__(self, param, parent_instance):
        assert isinstance(param, cb.EnumParameter)

        super().__init__(param, parent_instance)

        label = QLabel(param.label, self)
        self.grid.addWidget(label, 0, 0, 1, 1)

        self.cb = wrap_qwidget(QComboBox)()
        members = param.enum.__members__.values()
        label_attrib_name = "label" if hasattr(param.enum, "label") else "value"
        self.cb.addItems([getattr(e, label_attrib_name) for e in members])
        self.cb.currentIndexChanged.connect(self.__cb_event_handler)
        self.grid.addWidget(self.cb, 0, 1, 1, 1)

        self.update()

    def _update(self, *args, **kwargs):
        super()._update(*args, **kwargs)
        value = self._get_param_value()
        index = list(self.param.enum.__members__.values()).index(value)
        self.cb.setCurrentIndex(index)

    def __cb_event_handler(self, index):
        value = list(self.param.enum.__members__.values())[index]
        self._subwidget_event_handler(value)


class IntComboBoxPidget(PidgetStub):
    def __init__(self, param, parent_instance):
        assert isinstance(param, cb.IntParameter)

        super().__init__(param, parent_instance)

        label = QLabel(param.label, self)
        self.grid.addWidget(label, 0, 0, 1, 1)

        self.cb = wrap_qwidget(QComboBox)()
        members = param.valid_values
        self.cb.addItems([str(e) for e in members])
        self.cb.currentIndexChanged.connect(self.__cb_event_handler)
        self.grid.addWidget(self.cb, 0, 1, 1, 1)

        self.update()

    def _update(self, *args, **kwargs):
        super()._update(*args, **kwargs)
        value = self._get_param_value()
        index = self.param.valid_values.index(value)
        self.cb.setCurrentIndex(index)

    def __cb_event_handler(self, index):
        value = self.param.valid_values[index]
        self._subwidget_event_handler(value)


class BoolCheckboxPidget(PidgetStub):
    def __init__(self, param, parent_instance):
        super().__init__(param, parent_instance)

        self.checkbox = QCheckBox(param.label, self)
        self.checkbox.setTristate(False)
        self.checkbox.stateChanged.connect(self._subwidget_event_handler)
        self.grid.addWidget(self.checkbox, 0, 0, 1, 1)

        self.update()

    def _update(self, *args, **kwargs):
        super()._update(*args, **kwargs)
        value = self._get_param_value()
        self.checkbox.setChecked(value)


class IntSpinBoxPidget(PidgetStub):
    def __init__(self, param, parent_instance):
        assert isinstance(param, cb.IntParameter)

        super().__init__(param, parent_instance)

        self.grid.setColumnStretch(0, 7)
        self.grid.setColumnStretch(1, 3)

        label = QLabel(self)
        suffix = " [{}]".format(param.unit) if param.unit else ""
        label.setText(param.label + suffix)
        self.grid.addWidget(label, 0, 0, 1, 1)

        self.spin_box = wrap_qwidget(QSpinBox)(self)
        self.spin_box.setSingleStep(param.step)
        self.spin_box.setKeyboardTracking(False)
        self.spin_box.setRange(*(int(round(x)) for x in _limits_for_qt(param.limits)))

        if param.is_optional:
            self.checkbox = QCheckBox(param.optional_label, self)
            self.checkbox.setTristate(False)
            self.checkbox.stateChanged.connect(self.__checkbox_event_handler)

            self.grid.setColumnStretch(1, 1)
            self.grid.addWidget(self.checkbox, 0, 1, 1, 1)

            self.grid.setColumnStretch(2, 1)
            self.grid.addWidget(self.spin_box, 0, 2, 1, 1)

            self.spin_box.setValue(self.param.optional_default_set_value)
        else:
            self.checkbox = None
            self.grid.addWidget(self.spin_box, 0, 1, 1, 1)

        self.spin_box.valueChanged.connect(self.__spin_box_event_handler)

        self.update()

    def _update(self, *args, **kwargs):
        super()._update(*args, **kwargs)
        value = self._get_param_value()
        is_set = value is not None

        if is_set:
            self.spin_box.setValue(value)

        self.spin_box.setEnabled(is_set)

        if self.checkbox is not None:
            self.checkbox.setChecked(is_set)

    def __checkbox_event_handler(self, val):
        if val:
            val = self.spin_box.value()
        else:
            val = None

        self._subwidget_event_handler(val)

    def __spin_box_event_handler(self, val):
        if self.param.is_optional and not self.checkbox.isChecked():
            val = None

        self._subwidget_event_handler(val)


class FloatRangeSpinBoxesPidget(PidgetStub):
    def __init__(self, param, parent_instance):
        assert isinstance(param, cb.FloatRangeParameter)

        super().__init__(param, parent_instance)

        self.grid.setColumnStretch(0, 1)
        self.grid.setColumnStretch(1, 1)

        label = QLabel(self)
        suffix = " [{}]".format(param.unit) if param.unit else ""
        label.setText(param.label + suffix)
        self.grid.addWidget(label, 0, 0, 1, 1)

        self.spin_boxes = []
        for i in range(2):
            spin_box = wrap_qwidget(QDoubleSpinBox)(self)
            spin_box.setDecimals(param.decimals)
            spin_box.setSingleStep(10 ** (-param.decimals))
            spin_box.valueChanged.connect(lambda v, i=i: self.__spin_box_event_handler(v, i))
            spin_box.setKeyboardTracking(False)
            spin_box.setRange(*_limits_for_qt(param.limits))
            self.grid.addWidget(spin_box, 1, i, 1, 1)
            self.spin_boxes.append(spin_box)

        self.update()

    def _update(self, *args, **kwargs):
        super()._update(*args, **kwargs)
        values = self._get_param_value()

        for spin_box, val in zip(self.spin_boxes, values):
            spin_box.setValue(val)

    def __spin_box_event_handler(self, val, index):
        vals = self._get_param_value()
        vals[index] = val
        self._subwidget_event_handler(vals)


class FloatSpinBoxPidget(PidgetStub):
    def __init__(self, param, parent_instance):
        assert isinstance(param, cb.FloatParameter)

        super().__init__(param, parent_instance)

        self.grid.setColumnStretch(0, 7)
        self.grid.setColumnStretch(1, 3)

        label = QLabel(self)
        suffix = " [{}]".format(param.unit) if param.unit else ""
        label.setText(param.label + suffix)
        self.grid.addWidget(label, 0, 0, 1, 1)

        self.spin_box = wrap_qwidget(QDoubleSpinBox)(self)
        self.spin_box.setDecimals(param.decimals)
        self.spin_box.setSingleStep(10 ** (-param.decimals))
        self.spin_box.setKeyboardTracking(False)
        self.spin_box.setRange(*_limits_for_qt(param.limits))

        if param.is_optional:
            self.checkbox = QCheckBox(param.optional_label, self)
            self.checkbox.setTristate(False)
            self.checkbox.stateChanged.connect(self.__checkbox_event_handler)

            self.grid.setColumnStretch(1, 1)
            self.grid.addWidget(self.checkbox, 0, 1, 1, 1)

            self.grid.setColumnStretch(2, 1)
            self.grid.addWidget(self.spin_box, 0, 2, 1, 1)

            self.spin_box.setValue(self.param.optional_default_set_value)
        else:
            self.checkbox = None
            self.grid.addWidget(self.spin_box, 0, 1, 1, 1)

        self.spin_box.valueChanged.connect(self.__spin_box_event_handler)

        self.update()

    def _update(self, *args, **kwargs):
        super()._update(*args, **kwargs)
        value = self._get_param_value()
        is_set = value is not None

        if is_set:
            self.spin_box.setValue(value)

        self.spin_box.setEnabled(is_set)

        if self.checkbox is not None:
            self.checkbox.setChecked(is_set)

    def __checkbox_event_handler(self, val):
        if val:
            val = self.spin_box.value()
        else:
            val = None

        self._subwidget_event_handler(val)

    def __spin_box_event_handler(self, val):
        if self.param.is_optional and not self.checkbox.isChecked():
            val = None

        self._subwidget_event_handler(val)


class FloatSpinBoxAndSliderPidget(PidgetStub):
    NUM_SLIDER_STEPS = 200

    def __init__(self, param, parent_instance):
        assert isinstance(param, cb.FloatParameter)

        super().__init__(param, parent_instance)

        self.grid.setColumnStretch(0, 7)
        self.grid.setColumnStretch(1, 3)

        label = QLabel(self)
        suffix = " [{}]".format(param.unit) if param.unit else ""
        label.setText(param.label + suffix)
        self.grid.addWidget(label, 0, 0, 1, 1)

        self.spin_box = wrap_qwidget(QDoubleSpinBox)(self)
        self.spin_box.setDecimals(param.decimals)
        self.spin_box.setSingleStep(10 ** (-param.decimals))
        self.spin_box.valueChanged.connect(self._subwidget_event_handler)
        self.spin_box.setKeyboardTracking(False)
        self.spin_box.setRange(*_limits_for_qt(param.limits))
        self.grid.addWidget(self.spin_box, 0, 1, 1, 1)

        slider_widget = QWidget()
        slider_layout = QHBoxLayout(slider_widget)
        slider_layout.setContentsMargins(0, 0, 0, 0)
        slider_layout.addWidget(QLabel(str(param.limits[0])))
        self.slider = wrap_qwidget(QSlider)(QtCore.Qt.Horizontal)
        self.slider.setRange(0, self.NUM_SLIDER_STEPS)
        self.slider.sliderPressed.connect(self.__slider_event_handler)
        self.slider.valueChanged.connect(self.__slider_event_handler)
        slider_layout.addWidget(self.slider, 1)
        slider_layout.addWidget(QLabel(str(param.limits[1])))
        self.grid.addWidget(slider_widget, 1, 0, 1, 2)

        self.update()

    def _update(self, *args, **kwargs):
        super()._update(*args, **kwargs)
        value = self._get_param_value()
        self.spin_box.setValue(value)
        self.slider.setValue(self.__to_slider_scale(value))

    def __slider_event_handler(self, x=None):
        if x is None:
            x = self.slider.sliderPosition()

        self._subwidget_event_handler(self.__from_slider_scale(x))

    def __to_slider_scale(self, x):
        lower, upper = self.param.limits

        if self.param.logscale:
            lower = np.log(lower)
            upper = np.log(upper)
            x = np.log(x)

        y = (x - lower) / (upper - lower) * self.NUM_SLIDER_STEPS
        return int(round(y))

    def __from_slider_scale(self, y):
        lower, upper = self.param.limits

        if self.param.logscale:
            lower = np.log(lower)
            upper = np.log(upper)

        x = y / self.NUM_SLIDER_STEPS * (upper - lower) + lower

        if self.param.logscale:
            x = np.exp(x)

        return x



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import abc
import webbrowser
from typing import List, Optional

import qtawesome as qta

from PySide6.QtWidgets import QHBoxLayout, QLabel, QPushButton, QWidget

from acconeer.exptool.app.new._enums import ConnectionInterface
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.ui.icons import EXTERNAL_LINK, WARNING


class HintObject:
    def __init__(self, warning: str, tooltip: str, how_to_fix_url: str) -> None:
        self.warning = warning
        self.tooltip = tooltip
        self.how_to_fix_url = how_to_fix_url

    @staticmethod
    @abc.abstractmethod
    def _should_show(app_model: AppModel) -> bool:
        pass


class UserHintWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        app_model.sig_notify.connect(self._on_app_model_update)

        self._hints: List[HintObject] = []
        self._how_to_fix_url: Optional[str] = None

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.icon = qta.IconWidget()
        self.icon.setHidden(True)
        self.icon.setIcon(WARNING())
        layout.addWidget(self.icon)

        self.label = QLabel(self)
        self.label.setHidden(True)
        layout.addWidget(self.label)

        self.button = QPushButton(self)
        self.button.setIcon(EXTERNAL_LINK())
        self.button.setText("How to fix")
        self.button.clicked.connect(self._on_click)
        self.button.setHidden(True)
        layout.addWidget(self.button)

        self.setLayout(layout)

    def add_hint(self, hint: HintObject) -> None:
        self._hints.append(hint)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        for hint in self._hints:
            if hint._should_show(app_model):
                self.label.setText(hint.warning)
                self.label.setToolTip(hint.tooltip)
                self.icon.setToolTip(hint.tooltip)
                self._how_to_fix_url = hint.how_to_fix_url
                self.icon.setHidden(False)
                self.label.setHidden(False)
                if self._how_to_fix_url is not None:
                    self.button.setHidden(False)
                return

        self.icon.setHidden(True)
        self.label.setHidden(True)
        self.button.setHidden(True)

    def _on_click(self) -> None:
        if self._how_to_fix_url is not None:
            webbrowser.open_new_tab(self._how_to_fix_url)


class UnflashedDeviceHint(HintObject):
    def __init__(self) -> None:
        super().__init__(
            "Unflashed device",
            "The device needs to be flashed with exploration server firmware",
            r"https://docs.acconeer.com/en/latest/evk_setup/xc120_xe121.html",
        )

    @staticmethod
    def _should_show(app_model: AppModel) -> bool:
        if app_model.connection_interface not in [
            ConnectionInterface.SERIAL,
            ConnectionInterface.USB,
        ]:
            return False

        if (
            app_model.serial_connection_device is not None
            and app_model.connection_interface == ConnectionInterface.SERIAL
            and app_model.serial_connection_device.unflashed
        ):
            return True

        return (
            app_model.usb_connection_device is not None
            and app_model.connection_interface == ConnectionInterface.USB
            and app_model.usb_connection_device.unflashed
        )


class InaccessibleDeviceHint(HintObject):
    def __init__(self) -> None:
        super().__init__(
            "Device permissions",
            "The USB device permissions needs to be setup, "
            "update USB permissions or use Serial Port",
            r"https://docs.acconeer.com/en/latest/exploration_tool/"
            "installation_and_setup.html#linux-setup",
        )

    @staticmethod
    def _should_show(app_model: AppModel) -> bool:
        return (
            app_model.usb_connection_device is not None
            and app_model.connection_interface == ConnectionInterface.USB
            and not app_model.usb_connection_device.accessible
        )


class HintWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        hint_widget = UserHintWidget(app_model, self)
        layout.addWidget(hint_widget)

        # Prioritized hint order:
        # The first will have priority over the second
        # The second will have priority over the third...
        hint_widget.add_hint(InaccessibleDeviceHint())
        hint_widget.add_hint(UnflashedDeviceHint())

        self.setLayout(layout)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

from PySide6 import QtGui

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.depth_res = session_info["step_length_m"]
        self.smooth_max = et.utils.SmoothMax(sensor_config.update_rate)

    def setup(self, win):
        num_sensors = len(self.sensor_config.sensor)

        self.ampl_plot = win.addPlot(row=0, col=0, colspan=num_sensors)
        self.ampl_plot.setMenuEnabled(False)
        self.ampl_plot.setMouseEnabled(x=False, y=False)
        self.ampl_plot.hideButtons()
        self.ampl_plot.showGrid(x=True, y=True)
        self.ampl_plot.setLabel("bottom", "Depth (m)")
        self.ampl_plot.setLabel("left", "Amplitude")
        self.ampl_plot.setXRange(*self.depths.take((0, -1)))
        self.ampl_plot.addLegend(offset=(-10, 10))

        self.phase_plot = win.addPlot(row=1, col=0, colspan=num_sensors)
        self.phase_plot.setMenuEnabled(False)
        self.phase_plot.setMouseEnabled(x=False, y=False)
        self.phase_plot.hideButtons()
        self.phase_plot.showGrid(x=True, y=True)
        self.phase_plot.setLabel("bottom", "Depth (m)")
        self.phase_plot.setLabel("left", "Phase")
        self.phase_plot.setXRange(*self.depths.take((0, -1)))
        self.phase_plot.setYRange(-np.pi, np.pi)
        self.phase_plot.getAxis("left").setTicks(et.utils.pg_phase_ticks)

        self.ampl_curves = []
        self.phase_curves = []
        for i, sensor_id in enumerate(self.sensor_config.sensor):
            legend = "Sensor {}".format(sensor_id)
            pen = et.utils.pg_pen_cycler(i)
            ampl_curve = self.ampl_plot.plot(pen=pen, name=legend)
            phase_curve = self.phase_plot.plot(pen=pen)
            self.ampl_curves.append(ampl_curve)
            self.phase_curves.append(phase_curve)

        rate = self.sensor_config.update_rate
        xlabel = "Sweeps" if rate is None else "Time (s)"
        x_scale = 1.0 if rate is None else 1.0 / rate
        y_scale = self.depth_res
        x_offset = -self.processing_config.history_length * x_scale
        y_offset = self.depths[0] - 0.5 * self.depth_res
        is_single_sensor = len(self.sensor_config.sensor) == 1

        self.history_plots = []
        self.history_ims = []
        for i, sensor_id in enumerate(self.sensor_config.sensor):
            title = None if is_single_sensor else "Sensor {}".format(sensor_id)
            plot = win.addPlot(row=2, col=i, title=title)
            plot.setMenuEnabled(False)
            plot.setMouseEnabled(x=False, y=False)
            plot.hideButtons()
            plot.setLabel("bottom", xlabel)
            plot.setLabel("left", "Depth (m)")
            im = pg.ImageItem(autoDownsample=True)
            im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
            im.resetTransform()
            tr = QtGui.QTransform()
            tr.translate(x_offset, y_offset)
            tr.scale(x_scale, y_scale)
            im.setTransform(tr)
            plot.addItem(im)
            self.history_plots.append(plot)
            self.history_ims.append(im)

    def update(self, d):
        sweeps = d["data"]
        histories = np.abs(d["history"])

        ampls = np.abs(sweeps)

        for i, _ in enumerate(self.sensor_config.sensor):
            self.ampl_curves[i].setData(self.depths, ampls[i])
            self.phase_curves[i].setData(self.depths, np.angle(sweeps[i]))

            im = self.history_ims[i]
            history = histories[:, i]
            im.updateImage(history, levels=(0, 1.05 * history.max()))

        m = self.smooth_max.update(ampls)
        self.ampl_plot.setYRange(0, m)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtGui import QTransform
from PySide6.QtWidgets import QLabel, QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121, opser
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import PeakSortingMethod
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    HandledException,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    PluginState,
    PluginStateMessage,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components.a121 import SensorConfigEditor

from ._detector import (
    DetailedStatus,
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
    _load_algo_data,
)


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_ids: list[int] = attrs.field(factory=lambda: [1])
    config: DetectorConfig = attrs.field(factory=DetectorConfig)
    context: DetectorContext = attrs.field(factory=DetectorContext)


class PluginPresetId(Enum):
    DEFAULT = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], DetectorConfig]] = {
        PluginPresetId.DEFAULT.value: lambda: DetectorConfig()
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)
        self._detector_instance: Optional[Detector] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = DetectorConfig.from_json(file["config"][()])
        self.shared_state.context = opser.deserialize(file["context"], DetectorContext)

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            # Try to use the sensor ids from the last calibration
            if self.shared_state.context.sensor_ids:
                self.shared_state.sensor_ids = self.shared_state.context.sensor_ids

            for i in range(len(self.shared_state.sensor_ids)):
                if len(sensor_ids) > 0 and self.shared_state.sensor_ids[i] not in sensor_ids:
                    self.shared_state.sensor_ids[i] = sensor_ids[0]

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def update_sensor_ids(self, *, sensor_ids: list[int]) -> None:
        self.shared_state.sensor_ids = sensor_ids
        self.broadcast()

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())
        opser.serialize(self.shared_state.context, file.create_group("context"))

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config, context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.context = context
        self.shared_state.sensor_ids = list(next(iter(record.session_config.groups)).keys())

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._detector_instance = Detector(
            client=self.client,
            sensor_ids=self.shared_state.sensor_ids,
            detector_config=self.shared_state.config,
            context=self.shared_state.context,
        )
        self._detector_instance.start(recorder)
        self.callback(SetupMessage())

    def end_session(self) -> None:
        assert self._detector_instance
        if self._recorder is not None:
            self._recorder.close()
        self._detector_instance.stop()

    def get_next(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError

        assert self.client
        result = self._detector_instance.get_next()

        self.callback(backend.PlotMessage(result=result))

    @is_task
    def calibrate_detector(self) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError

        if not self.client.connected:
            raise RuntimeError

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

        try:
            self._detector_instance = Detector(
                client=self.client,
                sensor_ids=self.shared_state.sensor_ids,
                detector_config=self.shared_state.config,
                context=None,
            )
            self._detector_instance.calibrate_detector()
        except Exception as exc:
            msg = "Failed to calibrate detector"
            raise HandledException(msg) from exc
        finally:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        self.shared_state.context = self._detector_instance.context
        self.broadcast()


class PlotPlugin(PgPlotPlugin):
    _PLOT_HISTORY_FRAMES = 50

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[DetectorResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup()
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(multi_sensor_result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(self) -> None:
        self.plot_layout.clear()

        self.obst_vel_ys = [np.nan * np.ones(self._PLOT_HISTORY_FRAMES)]
        self.obst_dist_ys = [np.nan * np.ones(self._PLOT_HISTORY_FRAMES)]
        self.hist_x = np.linspace(
            -100, 0, self._PLOT_HISTORY_FRAMES
        )  # TODO: should no be hard coded

        # Upper row
        win = self.plot_layout
        win.setWindowTitle("Acconeer obstacle detector example")

        self.fftmap_plot = pg.PlotItem(title="FFT Map")
        self.fftmap_image = pg.ImageItem(autoDownsample=True)
        self.fftmap_image.setLookupTable(et.utils.pg_mpl_cmap("viridis"))

        self.fftmap_plot.setLabel("bottom", "Distance (m)")
        self.fftmap_plot.setLabel("left", "Velocity (cm/s)")
        self.fftmap_plot.addItem(self.fftmap_image)

        sublayout = win.addLayout(
            row=0,
            col=0,
            colspan=2,
        )
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.fftmap_plot, row=0, col=0)

        # Middle row

        self.bin0_plot = win.addPlot(
            row=1,
            col=0,
            title="Static objects (Zeroth FFT bin)",
        )

        self.bin0_plot.showGrid(x=True, y=True)
        self.bin0_plot.setLabel("bottom", "Range (cm)")
        self.bin0_plot.setLabel("left", "Amplitude")
        # self.bin0_plot.setXRange(0, 0)
        self.bin0_plot.addLegend()
        self.bin0_curves = [self.bin0_plot.plot(symbolSize=5, symbol="o")]

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.bin0_curve = self.bin0_plot.plot(**feat_kw)

        pen = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.bin0_threshold_curve = self.bin0_plot.plot(**feat_kw)

        bin0_plot_legend = pg.LegendItem(offset=(0.0, 0.45))
        bin0_plot_legend.setParentItem(self.bin0_plot)
        bin0_plot_legend.addItem(self.bin0_curve, "Sweep")
        bin0_plot_legend.addItem(self.bin0_threshold_curve, "Threshold")

        self.other_bins_plot = win.addPlot(
            row=1,
            col=1,
            title="Moving objects (Other FFT bins)",
        )

        self.other_bins_plot.showGrid(x=True, y=True)
        self.other_bins_plot.setLabel("bottom", "Range (cm)")
        self.other_bins_plot.setLabel("left", "Amplitude")
        # self.other_bins_plot.setXRange(0, 0)
        self.other_bins_plot.addLegend()
        self.other_bins_curves = [self.other_bins_plot.plot(symbolSize=5, symbol="o")]

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.other_bins_curve = self.other_bins_plot.plot(**feat_kw)

        pen = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.other_bins_threshold_curve = self.other_bins_plot.plot(**feat_kw)

        other_bins_plot_legend = pg.LegendItem(offset=(0.0, 0.45))
        other_bins_plot_legend.setParentItem(self.other_bins_plot)
        other_bins_plot_legend.addItem(self.other_bins_curve, "Sweep")
        other_bins_plot_legend.addItem(self.other_bins_threshold_curve, "Threshold")

        # Lower row

        self.angle_hist_plot = win.addPlot(
            row=2,
            col=0,
            title="Obstacle velocity history",
        )

        self.angle_hist_plot.showGrid(x=True, y=True)
        self.angle_hist_plot.setLabel("bottom", "Time (frames)")
        self.angle_hist_plot.setLabel("left", "Velocity (cm/s)")
        self.angle_hist_plot.setXRange(-100, 0)
        self.angle_hist_plot.addLegend()
        self.velocity_hist_curves = [self.angle_hist_plot.plot(symbolSize=5, symbol="o")]

        self.range_hist_plot = win.addPlot(
            row=2,
            col=1,
            title="Obstacle range history",
        )

        self.range_hist_plot.showGrid(x=True, y=True)
        self.range_hist_plot.setLabel("bottom", "Time (frames)")
        self.range_hist_plot.setLabel("left", "Range (cm)")
        self.range_hist_plot.setXRange(-100, 0)
        self.range_hist_plot.addLegend()
        self.range_hist_curves = [self.range_hist_plot.plot(symbolSize=5, symbol="o")]

        close_proximity_html = (
            '<p style="text-align: center; color: white; font-size: 10pt;">{}</p>'.format(
                "Close proximity detection",
            )
        )

        self.close_proximity_text_item = pg.TextItem(
            html=close_proximity_html,
            fill=pg.mkColor(0xFF, 0x7F, 0x0E),
            anchor=(0.5, 0),
        )
        self.range_hist_plot.addItem(self.close_proximity_text_item)
        self.close_proximity_text_item.hide()

    def draw_plot_job(self, multi_sensor_result: DetectorResult) -> None:
        # Get the first element as the plugin only supports single sensor operation.

        (pr,) = multi_sensor_result.processor_results.values()
        er = pr.subsweeps_extra_results[0]

        fftmap = er.fft_map
        fftmap_threshold = er.fft_map_threshold
        spf = fftmap.shape[0]

        transform = QTransform()
        transform.translate(
            100 * er.r[0], -100 * pr.extra_result.dv * spf / 2 - 0.5 * 100 * pr.extra_result.dv
        )
        transform.scale(100 * er.r[1] - 100 * er.r[0], 100 * pr.extra_result.dv)

        self.fftmap_image.setTransform(transform)
        self.fftmap_image.updateImage(
            np.fft.fftshift(fftmap, 0).T,
            levels=(0, 1.05 * np.max(fftmap)),
        )

        bin0 = fftmap[0, :]
        threshold_bin0 = fftmap_threshold[0, :]

        max_other_bins = np.max(fftmap[1:, :], axis=0)
        threshold_other_bins = fftmap_threshold[1, :]

        self.bin0_curve.setData(100 * er.r, bin0)
        self.bin0_threshold_curve.setData(100 * er.r, threshold_bin0)
        self.other_bins_curve.setData(100 * er.r, max_other_bins)
        self.other_bins_threshold_curve.setData(100 * er.r, threshold_other_bins)

        self.angle_hist_plot.setYRange(
            -100 * pr.extra_result.dv * spf / 2, 100 * pr.extra_result.dv * spf / 2
        )
        v = pr.targets[0].velocity if pr.targets else np.nan

        self.obst_vel_ys[0] = np.roll(self.obst_vel_ys[0], -1)
        self.obst_vel_ys[0][-1] = 100 * v  # m/s -> cm/s

        if np.isnan(self.obst_vel_ys[0]).all():
            self.velocity_hist_curves[0].setVisible(False)
        else:
            self.velocity_hist_curves[0].setVisible(True)
            self.velocity_hist_curves[0].setData(
                self.hist_x, self.obst_vel_ys[0], connect="finite"
            )

        min_range = 100 * er.r[0]
        max_range = 100 * er.r[-1]
        self.range_hist_plot.setYRange(min_range, max_range)
        r = pr.targets[0].distance if pr.targets else np.nan

        self.obst_dist_ys[0] = np.roll(self.obst_dist_ys[0], -1)
        self.obst_dist_ys[0][-1] = 100 * r  # m/s -> cm/s

        if np.isnan(self.obst_dist_ys[0]).all():
            self.range_hist_curves[0].setVisible(False)
        else:
            self.range_hist_curves[0].setVisible(True)
            self.range_hist_curves[0].setData(self.hist_x, self.obst_dist_ys[0], connect="finite")

        if multi_sensor_result.close_proximity_trig:
            close_prox = list(multi_sensor_result.close_proximity_trig.values())
            self.range_hist_plot.getAxis("left")
            self.close_proximity_text_item.setPos(-50, max_range)
            self.close_proximity_text_item.setVisible(close_prox[0])


class ViewPlugin(A121ViewPluginBase):
    sensor_config_editors: list[SensorConfigEditor]

    TEXT_MSG_MAP = {
        DetailedStatus.OK: "Ready to start.",
        DetailedStatus.END_LESSER_THAN_START: "'Range end' point must be greater than 'Range "
        + "start'.",
        DetailedStatus.CONTEXT_MISSING: "Run detector calibration.",
        DetailedStatus.CALIBRATION_MISSING: "Run detector calibration.",
        DetailedStatus.CONFIG_MISMATCH: (
            "Current configuration does not match the configuration "
            + "used during detector calibration. Run detector calibration."
        ),
    }

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self.app_model = app_model
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.calibrate_detector_button = QPushButton(icons.CALIBRATE(), "Calibrate detector")
        self.calibrate_detector_button.clicked.connect(self._on_calibrate_detector)

        self.defaults_button = QPushButton(icons.RESTORE(), "Reset settings and calibrations")
        self.defaults_button.clicked.connect(self._send_defaults_request)

        self.message_box = QLabel(self.sticky_widget)
        self.message_box.setWordWrap(True)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.calibrate_detector_button, 1, 0, 1, -1)
        button_group.layout().addWidget(self.defaults_button, 3, 0, 1, -1)
        button_group.layout().addWidget(self.message_box, 4, 0, 1, -1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor[DetectorConfig](
            title="Detector parameters",
            config_type=DetectorConfig,
            factory_mapping=self.get_pidget_mapping(),
            parent=self.scrolly_widget,
        )

        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "start_m": pidgets.FloatPidgetFactory(
                name_label_text="Range start:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "start_m"),
                suffix=" m",
                decimals=3,
            ),
            "end_m": pidgets.FloatPidgetFactory(
                name_label_text="Range end:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "end_m"),
                suffix=" m",
                decimals=3,
            ),
            "step_length": pidgets.IntPidgetFactory(
                name_label_text="Step length (in unit 2.5 mm):",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "step_length"),
                suffix=" ",
            ),
            "max_robot_speed": pidgets.FloatPidgetFactory(
                name_label_text="Max robot speed:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "max_robot_speed"),
                suffix=" m/s",
                limits=(0.001, None),
                decimals=3,
            ),
            "profile": pidgets.EnumPidgetFactory(
                name_label_text="Profile:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "profile"),
                enum_type=a121.Profile,
                label_mapping={
                    a121.Profile.PROFILE_1: "1 (best resolution)",
                    a121.Profile.PROFILE_2: "2",
                    a121.Profile.PROFILE_3: "3",
                    a121.Profile.PROFILE_4: "4",
                    a121.Profile.PROFILE_5: "5 (best SNR)",
                },
            ),
            "hwaas": pidgets.IntPidgetFactory(
                name_label_text="Hardware averaging (HWAAS):",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "hwaas"),
                suffix=" ",
            ),
            "num_std_threshold": pidgets.FloatPidgetFactory(
                name_label_text="Noise threshold:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "num_std_threshold"),
                suffix=" ",
                decimals=3,
            ),
            "num_mean_threshold": pidgets.FloatPidgetFactory(
                name_label_text="Mean threshold:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "num_mean_threshold"),
                suffix=" ",
                decimals=3,
            ),
            "num_frames_in_recorded_threshold": pidgets.IntPidgetFactory(
                name_label_text="Nbr frames used to determine treshold:",
                name_label_tooltip=get_attribute_docstring(
                    DetectorConfig, "num_frames_in_recorded_threshold"
                ),
                suffix=" ",
            ),
            "sweeps_per_frame": pidgets.IntPidgetFactory(
                name_label_text="Number of sweeps per frame:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "sweeps_per_frame"),
                suffix=" ",
            ),
            "update_rate": pidgets.FloatPidgetFactory(
                name_label_text="Detector update rate:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "update_rate"),
                suffix=" Hz",
                decimals=1,
            ),
            "peak_sorting_method": pidgets.EnumPidgetFactory(
                name_label_text="Peak sorting method:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "peak_sorting_method"),
                enum_type=PeakSortingMethod,
                label_mapping={
                    PeakSortingMethod.CLOSEST: "Closest",
                    PeakSortingMethod.STRONGEST: "Strongest",
                },
            ),
            "dead_reckoning_duration_s": pidgets.FloatPidgetFactory(
                name_label_text="Dead reckoning duration:",
                name_label_tooltip=get_attribute_docstring(
                    DetectorConfig, "dead_reckoning_duration_s"
                ),
                suffix=" s",
                decimals=1,
            ),
            "kalman_sensitivity": pidgets.FloatSliderPidgetFactory(
                name_label_text="Kalman filter sensitivity:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "kalman_sensitivity"),
                decimals=2,
                limits=(0.001, 1),
                show_limit_values=False,
                limit_texts=("Higher robustness", "More Responsive"),
            ),
            "enable_close_proximity_detection": pidgets.CheckboxPidgetFactory(
                name_label_text="Enable close proximity detection",
                name_label_tooltip=get_attribute_docstring(
                    DetectorConfig, "enable_close_proximity_detection"
                ),
            ),
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.message_box.setText("")
        else:
            self.config_editor.set_data(state.config)

            detector_status = Detector.get_detector_status(
                state.config, state.context, state.sensor_ids
            )

            self.message_box.setText(self.TEXT_MSG_MAP[detector_status.detector_state])

            validation_results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(validation_results)
            not_handled = self.misc_error_view.handle_validation_results(not_handled)
            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[
                self.defaults_button,
                self.config_editor,
                self.sensor_id_pidget,
            ],
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

        state = app_model.backend_plugin_state

        if state is None:
            ready_to_start = False
            config_valid = False
        else:
            detector_status = Detector.get_detector_status(
                state.config, state.context, state.sensor_ids
            )

            ready_to_start = detector_status.ready_to_start
            config_valid = self._config_valid(state) and self.config_editor.is_ready

        self.calibrate_detector_button.setEnabled(
            visual_policies.start_button_enabled(app_model, extra_condition=config_valid),
        )
        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=ready_to_start and config_valid
            )
        )

    def _config_valid(self, state: SharedState) -> bool:
        try:
            state.config.validate()
        except a121.ValidationResult:
            return False
        else:
            return True

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_ids.rpc(self.app_model.put_task, sensor_ids=[sensor_id])

    def _on_config_update(self, config: DetectorConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_calibrate_detector(self) -> None:
        BackendPlugin.calibrate_detector.rpc(self.app_model.put_task)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


OBSTACLE_DETECTOR_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="obstacle_detector",
    title="Obstacle detection",
    docs_link="https://docs.acconeer.com/en/latest/example_apps/a121/obstacle_detection.html",
    description="Measure distance and angle to objects from a moving platform.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py

from PySide6.QtWidgets import QPushButton, QVBoxLayout

from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo._utils import APPROX_BASE_STEP_LENGTH_M
from acconeer.exptool.a121.algo.vibration import (
    ReportedDisplacement,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components.a121 import RangeHelpView

from ._configs import get_high_frequency_config, get_low_frequency_config
from ._example_app import ExampleApp, ExampleAppConfig, ExampleAppResult, _load_algo_data
from .plot import VibrationPlot


log = logging.getLogger(__name__)


class PluginPresetId(Enum):
    LOW_FREQ = auto()
    HIGH_FREQ = auto()


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: ExampleAppConfig = attrs.field(factory=ExampleAppConfig)


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    example_app_config: ExampleAppConfig
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], ExampleAppConfig]] = {
        PluginPresetId.LOW_FREQ.value: lambda: get_low_frequency_config(),
        PluginPresetId.HIGH_FREQ.value: lambda: get_high_frequency_config(),
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._exempel_app_instance: Optional[ExampleApp] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = ExampleAppConfig.from_json(file["config"][()])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: ExampleAppConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.sensor_id = record.sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._example_app_instance = ExampleApp(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            example_app_config=self.shared_state.config,
        )
        self._example_app_instance.start(recorder)
        self.callback(SetupMessage(example_app_config=self.shared_state.config))

    def end_session(self) -> None:
        if self._example_app_instance is None:
            raise RuntimeError
        if self._recorder is not None:
            self._recorder.close()
        self._example_app_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._example_app_instance is None:
            raise RuntimeError
        result = self._example_app_instance.get_next()

        self.callback(backend.PlotMessage(result=result))


class ViewPlugin(A121ViewPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.range_helper = RangeHelpView()
        scrolly_layout.addWidget(self.range_helper)

        self.sensor_config_editor = AttrsConfigEditor(
            title="Sensor parameters",
            factory_mapping=self._get_sensor_pidget_mapping(),
            config_type=ExampleAppConfig,
            parent=self.scrolly_widget,
        )
        self.sensor_config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.sensor_config_editor)

        self.config_editor = AttrsConfigEditor(
            title="Example app parameters",
            factory_mapping=self._get_pidget_mapping(),
            config_type=ExampleAppConfig,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_sensor_pidget_mapping(cls) -> PidgetFactoryMapping:
        # Note: Incomplete mapping
        return {
            "measured_point": pidgets.IntPidgetFactory(
                name_label_text="Measured point:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "measured_point"),
                limits=(0, None),
            ),
            "profile": pidgets.EnumPidgetFactory(
                name_label_text="Profile:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "profile"),
                enum_type=a121.Profile,
                label_mapping={
                    a121.Profile.PROFILE_1: "1 (shortest)",
                    a121.Profile.PROFILE_2: "2",
                    a121.Profile.PROFILE_3: "3",
                    a121.Profile.PROFILE_4: "4",
                    a121.Profile.PROFILE_5: "5 (longest)",
                },
            ),
            "frame_rate": pidgets.OptionalFloatPidgetFactory(
                name_label_text="Frame rate:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "frame_rate"),
                checkbox_label_text="Limit",
                suffix=" Hz",
                decimals=1,
                limits=(1, None),
                init_set_value=10,
            ),
            "sweep_rate": pidgets.FloatPidgetFactory(
                name_label_text="Sweep rate:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "sweep_rate"),
                suffix=" Hz",
                decimals=1,
                limits=(1, None),
            ),
            "sweeps_per_frame": pidgets.IntPidgetFactory(
                name_label_text="Sweeps per frame:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "sweeps_per_frame"),
                limits=(1, 2048),
            ),
            "hwaas": pidgets.IntPidgetFactory(
                name_label_text="HWAAS:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "hwaas"),
                limits=(1, 511),
            ),
            "double_buffering": pidgets.CheckboxPidgetFactory(
                name_label_text="Double buffering",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "double_buffering"),
            ),
            "continuous_sweep_mode": pidgets.CheckboxPidgetFactory(
                name_label_text="Continuos sweep mode",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "continuous_sweep_mode"
                ),
            ),
            "inter_frame_idle_state": pidgets.EnumPidgetFactory(
                enum_type=a121.IdleState,
                name_label_text="Inter frame idle state:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "inter_frame_idle_state"
                ),
                label_mapping={
                    a121.IdleState.DEEP_SLEEP: "Deep sleep",
                    a121.IdleState.SLEEP: "Sleep",
                    a121.IdleState.READY: "Ready",
                },
            ),
            "inter_sweep_idle_state": pidgets.EnumPidgetFactory(
                enum_type=a121.IdleState,
                name_label_text="Inter sweep idle state:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "inter_sweep_idle_state"
                ),
                label_mapping={
                    a121.IdleState.DEEP_SLEEP: "Deep sleep",
                    a121.IdleState.SLEEP: "Sleep",
                    a121.IdleState.READY: "Ready",
                },
            ),
        }

    @classmethod
    def _get_pidget_mapping(cls) -> PidgetFactoryMapping:
        # Note: Incomplete mapping
        return {
            "time_series_length": pidgets.IntPidgetFactory(
                name_label_text="Time series length:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "time_series_length"),
                limits=(0, None),
            ),
            "lp_coeff": pidgets.FloatSliderPidgetFactory(
                name_label_text="Time filtering coefficient:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "lp_coeff"),
                suffix="",
                limits=(0, 1),
                decimals=2,
            ),
            "threshold_margin": pidgets.FloatSliderPidgetFactory(
                name_label_text="Threshold margin:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "threshold_margin"),
                suffix="um",
                limits=(0, 100),
                decimals=1,
            ),
            "amplitude_threshold": pidgets.FloatPidgetFactory(
                name_label_text="Amplitude threshold:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "amplitude_threshold"
                ),
                decimals=0,
                limits=(0, None),
            ),
            "reported_displacement_mode": pidgets.EnumPidgetFactory(
                name_label_text="Displacement mode:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "reported_displacement_mode"
                ),
                enum_type=ReportedDisplacement,
                label_mapping={
                    ReportedDisplacement.AMPLITUDE: "Amplitude",
                    ReportedDisplacement.PEAK2PEAK: "Peak to peak",
                },
            ),
            "low_frequency_enhancement": pidgets.CheckboxPidgetFactory(
                name_label_text="Low frequency enhancement",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "low_frequency_enhancement"
                ),
            ),
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.sensor_config_editor.set_data(None)
        else:
            self.config_editor.set_data(state.config)
            self.sensor_config_editor.set_data(state.config)
            self.sensor_id_pidget.set_data(state.sensor_id)

            results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

            not_handled = self.sensor_config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

            self.range_helper.set_data(ExampleApp._get_sensor_config(state.config).subsweeps[0])

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[self.config_editor, self.sensor_id_pidget],
        )

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[self.sensor_config_editor, self.sensor_id_pidget],
        )

        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model,
                extra_condition=self.config_editor.is_ready and self.sensor_config_editor.is_ready,
            )
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

    def _on_config_update(self, config: ExampleAppConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[ExampleAppResult] = None
        self._is_setup = False
        self._vibration_plot = VibrationPlot()

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self._setup(example_app_config=message.example_app_config)
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self._draw_plot_job(example_app_result=self._plot_job)
        finally:
            self._plot_job = None

    def _setup(self, example_app_config: ExampleAppConfig) -> None:
        self.plot_layout.clear()
        meas_dist_m = example_app_config.measured_point * APPROX_BASE_STEP_LENGTH_M
        self._vibration_plot.setup_plot(self.plot_layout, meas_dist_m)

    def _draw_plot_job(self, example_app_result: ExampleAppResult) -> None:
        self._vibration_plot.update_plot(
            result=example_app_result,
            extra_result=example_app_result.processor_extra_result,
        )


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


VIBRATION_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="vibration",
    title="Vibration measurement",
    docs_link="https://docs.acconeer.com/en/latest/example_apps/a121/vibration.html",
    description="Quantify the frequency content of vibrating object.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Low frequency", preset_id=PluginPresetId.LOW_FREQ),
        PluginPresetBase(name="High frequency", preset_id=PluginPresetId.HIGH_FREQ),
    ],
    default_preset_id=PluginPresetId.LOW_FREQ,
)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved
from __future__ import annotations

import enum
import time
from typing import Dict, List, Optional

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool._core.class_creation.attrs import attrs_dict_ndarray_isclose
from acconeer.exptool.a121.algo import AlgoProcessorConfigBase
from acconeer.exptool.a121.algo.distance import DetectorResult


TIME_HISTORY_S = 30
UPDATES_PER_SECOND = 4
CLOSE_RANGE_START = 0.07


class ProcessorLevelStatus(enum.Enum):
    """Status from the processor"""

    IN_RANGE = enum.auto()
    NO_DETECTION = enum.auto()
    OVERFLOW = enum.auto()
    OUT_OF_RANGE = enum.auto()


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    level_and_time_for_plotting: Dict[str, npt.NDArray[np.float64]] = attrs.field(
        eq=attrs_dict_ndarray_isclose
    )


@attrs.frozen(kw_only=True)
class ProcessorResult:
    """Processor results"""

    peak_detected: Optional[bool]
    peak_status: Optional[ProcessorLevelStatus]
    filtered_level: Optional[float]
    extra_result: ProcessorExtraResult


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    median_filter_length: int = attrs.field(default=5)
    """Length of the median filter used to stabilize the level result."""
    num_medians_to_average: int = attrs.field(default=5)
    """Number of median values averaged to obtain the level result."""
    tank_range_start_m: float = attrs.field(default=0.030)
    """Minimum detection distance."""
    tank_range_end_m: float = attrs.field(default=0.50)
    """Maximum detection distance."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        return []


class Processor:
    """Tank level indicator processor
    :param processor_config: Processor configuration.
    """

    def __init__(
        self,
        processor_config: ProcessorConfig,
    ):
        self.tank_range_end_m = processor_config.tank_range_end_m
        self.tank_range_start_m = processor_config.tank_range_start_m
        self.median_filter_length = processor_config.median_filter_length
        self.num_medians_to_average = processor_config.num_medians_to_average

        assert self.tank_range_start_m < self.tank_range_end_m

        self.median_counter = 0
        self.mean_counter = 0

        self.level_history = np.full(self.median_filter_length, np.nan)
        self.median_vector = np.full(self.num_medians_to_average, np.nan)

        self.near_edge_status_list: List[Optional[bool]] = []
        self.peak_status_list: List[bool] = []

        self.level_and_time_for_plotting = {
            "time": np.full((TIME_HISTORY_S * UPDATES_PER_SECOND,), np.nan),
            "level": np.full((TIME_HISTORY_S * UPDATES_PER_SECOND,), np.nan),
        }

        self.start_time = time.time()

    def _update_level_history(self, result: DetectorResult) -> None:
        self.median_counter += 1
        assert result.distances is not None

        if len(result.distances) != 0:
            self.level_history = np.append(
                self.level_history, (self.tank_range_end_m - result.distances[0])
            )
        else:
            self.level_history = np.append(self.level_history, np.nan)
        self.level_history = self.level_history[1:]

        self.near_edge_status_list.append(result.near_edge_status)

        if self.median_counter == self.median_filter_length:
            self.median_counter = 0
            # store level in a vector of medians
            self.median_vector = np.append(self.median_vector, np.median(self.level_history))[1:]
            # store peak status from near_edge_status_list to peak_status_list
            self.peak_status_list.append(
                self.near_edge_status_list.count(True) > len(self.near_edge_status_list) / 2
            )
            # clear the edge status list
            self.near_edge_status_list.clear()
            self.mean_counter += 1

    def _update_level_and_time_for_plotting(
        self, filtered_level: Optional[float], rel_time: float
    ) -> None:
        if filtered_level is not None:
            self.level_and_time_for_plotting["time"] = (
                np.append(self.level_and_time_for_plotting["time"], 0) - rel_time
            )[1:]
            self.level_and_time_for_plotting["level"] = (
                np.append(self.level_and_time_for_plotting["level"], filtered_level)
            )[1:]

    def process(
        self, detector_result: Dict[int, DetectorResult], detector_start_m: float
    ) -> ProcessorResult:
        # Get the first detector result (single sensor operation).
        (result,) = list(detector_result.values())

        # Assign level status
        # peak_detected: True if a peak is detected by the distance detector in n consecutive
        # frames, where n=median_filter_length.
        # peak_status: Assigned based on the location of the peak in relation to the detection
        # range defined by tank_start and tank_end.
        self._update_level_history(result)

        peak_detected = None
        peak_status = None
        filtered_level = None

        if self.mean_counter == self.num_medians_to_average:
            self.mean_counter = 0
            # assign filtered_level every num_medians_to_average samples
            if np.any(~np.isnan(self.median_vector)):
                filtered_level = np.nanmean(self.median_vector)
            else:
                filtered_level = np.nan

            # assign the detection and peak status
            peak_detected = False
            peak_status = ProcessorLevelStatus.NO_DETECTION

            if ~np.isnan(filtered_level):
                peak_detected = True
                if filtered_level < 0:
                    peak_status = ProcessorLevelStatus.OUT_OF_RANGE
                elif filtered_level > self.tank_range_end_m - self.tank_range_start_m or (
                    detector_start_m <= self.tank_range_start_m  # relevant for level tracking
                    and self.peak_status_list.count(True) > len(self.peak_status_list) / 2
                    and self.tank_range_start_m >= CLOSE_RANGE_START
                ):
                    peak_status = ProcessorLevelStatus.OVERFLOW
                else:
                    peak_status = ProcessorLevelStatus.IN_RANGE
            else:
                if (
                    detector_start_m <= self.tank_range_start_m  # relevant for level tracking
                    and self.peak_status_list.count(True) > len(self.peak_status_list) / 2
                    and self.tank_range_start_m >= CLOSE_RANGE_START
                ):
                    peak_status = ProcessorLevelStatus.OVERFLOW

            if peak_status in (ProcessorLevelStatus.OVERFLOW, ProcessorLevelStatus.OUT_OF_RANGE):
                filtered_level = np.nan

            rel_time = time.time() - self.start_time
            if rel_time >= 1 / UPDATES_PER_SECOND:
                self._update_level_and_time_for_plotting(filtered_level, rel_time)
                self.start_time = time.time()

            # clear the peak status list
            self.peak_status_list.clear()

        # extra results for plotting
        extra_result = ProcessorExtraResult(
            level_and_time_for_plotting=self.level_and_time_for_plotting
        )

        return ProcessorResult(
            peak_detected=peak_detected,
            peak_status=peak_status,
            extra_result=extra_result,
            filtered_level=filtered_level,
        )



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved
from __future__ import annotations

import typing as t

from . import core


class RegistryPersistor(core.Persistor):
    """
    Persistor implementation that can have other persistors registered to it.

    Its 'save' and 'load' implementations exclusively call registered persistor's
    'save' and 'load'.

    This persistor is the hub for recursion and allows circumventing
    the combinatorial explosion of the problem.
    """

    _REGISTRY: t.ClassVar[t.Dict[str, t.Type[core.Persistor]]] = {}

    @classmethod
    def register_persistor(cls, __persistor: t.Type[core.Persistor]) -> t.Type[core.Persistor]:
        """
        Registers a persistor that can extend the handled types.

        This can be called many times with the same persistor without repercussions.
        """
        cls._REGISTRY[__persistor.__name__] = __persistor
        return __persistor

    @classmethod
    def priority_higher_than(cls, __persistor: t.Type[core.Persistor]) -> int:
        """Returns a priority that is higher than the priority of the specified persistor."""
        try:
            return cls._REGISTRY[__persistor.__name__].PRIORITY + 1
        except KeyError:
            msg = f"Persistor {__persistor} is not in the registry"
            raise RuntimeError(msg)

    @classmethod
    def _get_applicable_persistors(cls, __type: type) -> t.List[t.Type[core.Persistor]]:
        """Retrieves the persistors that can handle the specified type"""
        return sorted(
            (persistor for persistor in cls._REGISTRY.values() if persistor.is_applicable(__type)),
            key=lambda p: p.PRIORITY,
            reverse=True,
        )

    @classmethod
    def registry_size(cls) -> int:
        return len(cls._REGISTRY)

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return len(cls._get_applicable_persistors(__type)) > 0

    def _save(self, instance: t.Any) -> None:
        persistor_classes = self._get_applicable_persistors(self.type_tree.data)

        if not persistor_classes:
            msg = f"No applicable persistors for instance {instance!r:.100}"
            raise core.SaveError(msg)

        errors = []
        for persistor_class in persistor_classes:
            try:
                persistor_class(self.parent_group, self.name, self.type_tree).save(instance)
            except core.SaveError as save_error:
                errors += [save_error]
                continue
            else:
                return

        msg = f"Could not save instance of type '{type(instance)}' (with expected type {self.type_tree.data}) using any of the persistors {persistor_classes}."
        raise core.SaveErrorGroup(
            msg,
            errors,
        )

    def _load(self) -> t.Any:
        persistor_classes = self._get_applicable_persistors(self.type_tree.data)

        if not persistor_classes:
            msg = f"No applicable persistors for expected type {self.type_tree.data}"
            raise core.LoadError(msg)

        errors = []
        for persistor_class in persistor_classes:
            try:
                res = persistor_class(self.parent_group, self.name, self.type_tree).load()
            except core.LoadError as load_error:
                errors += [load_error]
                continue
            else:
                return res

        msg = f"Could not load an instance of type {self.type_tree.data} with any of the persistors {persistor_classes}."
        raise core.LoadErrorGroup(
            msg,
            errors,
        )



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import json
import logging
import typing as t
from pathlib import Path

import attrs
import typing_extensions as te
from packaging.version import Version

from PySide6.QtCore import QSize, Qt, Signal
from PySide6.QtGui import QAction, QIcon, QMouseEvent
from PySide6.QtWidgets import (
    QBoxLayout,
    QCheckBox,
    QDockWidget,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QToolBar,
    QToolButton,
    QWidget,
)

import acconeer.exptool
from acconeer.exptool import a121
from acconeer.exptool.a121.algo import distance, presence
from acconeer.exptool.app.new.storage import get_config_dir
from acconeer.exptool.app.new.ui.icons import CHART_BAR, CHART_LINE, CLOSE, COG, EDIT, INFO, MEMORY
from acconeer.exptool.utils import get_module_version

from .animation import run_blink_animation
from .event_system import ChangeIdEvent, EventBroker
from .services import (
    distance_config_input,
    memory_breakdown_output,
    power_consumption_vs_rate_output,
    power_curve_output,
    presence_config_input,
    session_config_input,
)


_QWidgetT = t.TypeVar("_QWidgetT", bound=t.Type[QWidget])


log = logging.getLogger(__name__)

_ALLOWED_DOCK_AREA = Qt.DockWidgetArea.BottomDockWidgetArea

_DATA_IS_APPROXIMATE_DESCRIPTION = "<br><br>".join(
    [
        "All data and numbers presented in this tab are approximations based on RSS A121-v1.12.0, <b>not measurements</b>.",
        "The goal is to let you experiment with configurations and do comparisons.",
        "Use this tool as a way to quickly narrow down which configuration could satisfy your "
        + "constraints before you do measurements to get real numbers.",
    ]
)


_TAB_BRIEF_DESCRIPTION = "<br><br>".join(
    [
        "This tab allows you to experiment and visualize resource "
        + "requirements for different configurations. ",
        'The top bar is a palette with all the "blocks" you can use. '
        + "Click on them to add an additional block of that type.",
        "Some blocks have an id in [brackets] in the title bar. "
        + "This is used to connect inputs and outputs.",
        "You can hover a block's info icon for a longer description "
        + "of what the block does and how you may use it.",
    ]
)


@attrs.mutable
class _UserUnderstandings:
    """
    Simple class that keeps track of what the user understands.

    Used to control pop-ups with information when opening the resource tab.
    """

    _PATH: t.ClassVar[Path] = get_config_dir() / "user_understandings.json"

    data_is_approximate: bool
    tab_brief: bool  # Not used anymore
    lib_version: str

    @classmethod
    def require(cls) -> te.Self:
        """Returns a persisted instance or creates a new instance"""
        try:
            with cls._PATH.open("r") as file:
                cached = cls(**json.load(file))

                if Version(cached.lib_version) < Version("v7.6.1"):
                    # v7.6.1 slightly modified the approximate disclaimer text to
                    # emphasize that the numbers displayed are not measurements.
                    cached.data_is_approximate = False

                return cached
        except (FileNotFoundError, json.JSONDecodeError):
            return cls(
                data_is_approximate=False,
                tab_brief=False,
                lib_version=get_module_version(acconeer.exptool),
            )

    def save(self) -> None:
        with self._PATH.open("w") as file:
            return json.dump(attrs.asdict(self), file)


class _TitleBarLabel(QLabel):
    def __init__(self, text: str, parent: t.Optional[QWidget] = None) -> None:
        super().__init__(text=text, parent=parent)
        self.setTextFormat(Qt.TextFormat.RichText)

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        event.ignore()


class _EditableTitleBarLabel(QWidget):
    INTERESTS: t.ClassVar[set[type]] = {ChangeIdEvent}
    description: t.ClassVar[str] = ""
    window_title: str = ""

    def __init__(self, text: str, broker: EventBroker, parent: t.Optional[QWidget] = None) -> None:
        super().__init__(parent=parent)
        self._broker = broker

        layout = QHBoxLayout(self)

        self.label = _TitleBarLabel(text, self)
        layout.addWidget(self.label)

        edit_button = QPushButton()
        edit_button.setIcon(EDIT())
        edit_button.clicked.connect(self._on_edit_title)
        layout.addWidget(edit_button)

        self.setLayout(layout)

        broker.install_service(self)
        self.uninstall_function = lambda: broker.uninstall_service(self)

    def _on_edit_title(self) -> None:
        text, ok = QInputDialog.getText(self, "Edit title", "Title:", text=self.text())

        if ok and text and text != self.text():
            text = self._broker.change_id(text, self.text())
            self._broker.offer_event(ChangeIdEvent(self.text(), text))

    def text(self) -> str:
        return self.label.text()[3:-3]

    def handle_event(self, event: t.Any) -> None:
        if isinstance(event, ChangeIdEvent):
            self._handle_change_id_event(event)
        else:
            raise NotImplementedError

    def _handle_change_id_event(self, event: ChangeIdEvent) -> None:
        if self.text() == event.old_id:
            self.label.setText(f"<b>{event.new_id}<\b>")

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        event.ignore()


class _TitleBarWidget(QWidget):
    sig_exit_clicked = Signal()

    def __init__(
        self,
        title: str,
        tooltip: str,
        parent: QWidget,
    ) -> None:
        super().__init__(parent=parent)
        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground)
        self.setObjectName("_TitleBarWidget")
        self.setStyleSheet("#_TitleBarWidget { background: #ebebeb; }")

        fixed_label = _TitleBarLabel(title)

        info_icon = QToolButton()
        info_icon.setIcon(INFO())
        info_icon.setIconSize(QSize(16, 16))
        info_icon.setDisabled(True)
        info_icon.setToolTip(tooltip)

        close_button = QToolButton()
        close_button.setIcon(CLOSE())
        close_button.clicked.connect(self.sig_exit_clicked.emit)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(fixed_label, stretch=0)
        layout.addWidget(info_icon, stretch=0, alignment=Qt.AlignmentFlag.AlignLeft)
        layout.addWidget(close_button, stretch=1, alignment=Qt.AlignmentFlag.AlignRight)

        self.setLayout(layout)

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        event.ignore()


class _EditableTitleBarWidget(_TitleBarWidget):
    def __init__(
        self,
        editable_title: t.Optional[str],
        fixed_title: str,
        tooltip: str,
        broker: EventBroker,
        parent: QWidget,
    ) -> None:
        super().__init__(fixed_title, tooltip, parent)

        if editable_title is not None:
            self.editable_label = _EditableTitleBarLabel(f"<b>{editable_title}<\b>", broker)
            layout = self.layout()
            if isinstance(layout, QBoxLayout):
                layout.insertWidget(1, self.editable_label, stretch=0)


@attrs.frozen
class _SpawnInputEvent:
    compare_configs: bool


class _DockWidget(QDockWidget):
    def __init__(
        self,
        title: str,
        tooltip: str,
        widget: QWidget,
        close_hook: t.Callable[[], None],
        parent: QWidget,
    ) -> None:
        super().__init__(parent=parent)
        self.title = title
        self.tooltip = tooltip
        self.setWindowTitle(title)

        self.setWidget(widget)
        self.setAllowedAreas(_ALLOWED_DOCK_AREA)
        self._close_hook = close_hook

        self._set_title_bar_widget()

    def _set_title_bar_widget(self) -> None:
        title_bar = _TitleBarWidget(self.title, self.tooltip, self)
        title_bar.sig_exit_clicked.connect(self.close)
        self.setTitleBarWidget(title_bar)

    def close(self) -> bool:
        self._close_hook()
        return super().close()


class _InputDockWidget(_DockWidget):
    INTERESTS: t.ClassVar[set[type]] = {_SpawnInputEvent, ChangeIdEvent}
    description: t.ClassVar[str] = ""

    def __init__(
        self,
        id_: str,
        fixed_title: str,
        tooltip: str,
        widget: QWidget,
        close_hook: t.Callable[[], None],
        broker: EventBroker,
        parent: QWidget,
    ) -> None:
        self._id = id_
        self._fixed_title = fixed_title
        self._broker = broker
        super().__init__(fixed_title, tooltip, widget, close_hook, parent=parent)
        self.window_title = f"{self._fixed_title} - {self._id}"
        broker.install_service(self)
        self.uninstall_function = lambda: broker.uninstall_service(self)

    def _set_title_bar_widget(self) -> None:
        title_bar = _EditableTitleBarWidget(
            self._id, self._fixed_title, self.tooltip, self._broker, self
        )
        title_bar.sig_exit_clicked.connect(self.close)
        self.setTitleBarWidget(title_bar)

    def handle_event(self, event: t.Any) -> None:
        if isinstance(event, _SpawnInputEvent) and not event.compare_configs:
            self.close()
        if isinstance(event, ChangeIdEvent) and self._id == event.old_id:
            self._id = event.new_id
            self.window_title = f"{self._fixed_title} - {self._id}"

    def close(self) -> bool:
        self.uninstall_function()
        return super().close()


class ResourceMainWidget(QMainWindow):
    def __init__(self) -> None:
        super().__init__()

        self._broker = EventBroker()

        # default is AnimatedDocks | AllowTabbedDocks
        self.setDockOptions(
            QMainWindow.DockOption.AllowNestedDocks | QMainWindow.DockOption.AnimatedDocks
        )

        toolbar = QToolBar()
        toolbar.setFixedHeight(64)
        toolbar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        toolbar.setIconSize(QSize(24, 24))
        toolbar.setFloatable(False)
        toolbar.setMovable(False)

        self._compare_cb = QCheckBox("Compare")
        self._compare_cb.setToolTip(
            "Allow to open multiple configurations within the window to compare plots and numbers."
        )
        self._compare_cb.setChecked(False)

        self._populate_palette(toolbar)
        self.setCentralWidget(toolbar)

        # Default loadout
        power_curve = self._create_power_curve_output()
        self._add_dock_widget(power_curve)
        self._add_dock_widget_below_target(
            target=power_curve,
            new_widget=self._create_memory_breakdown_output(),
        )
        self._add_dock_widget_right_of_target(
            target=power_curve,
            new_widget=self._create_power_consumption_vs_rate_output(),
        )
        self.spawn_input_block(a121.SessionConfig(), animate=False)

    def showEvent(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().showEvent(*args, **kwargs)

        user_understandings = _UserUnderstandings.require()

        if not user_understandings.data_is_approximate:
            disclaimer_box = QMessageBox(
                QMessageBox.Icon.Information,
                "Data is approximate",
                _DATA_IS_APPROXIMATE_DESCRIPTION,
                parent=self,
            )
            understand_cb = QCheckBox("I understand", self)
            disclaimer_box.setCheckBox(understand_cb)
            disclaimer_box.exec()
            user_understandings.data_is_approximate = understand_cb.isChecked()

        user_understandings.save()

    def spawn_input_block(self, config: t.Any, animate: bool = True) -> None:
        self._broker.offer_event(_SpawnInputEvent(self._compare_cb.isChecked()))

        if isinstance(config, a121.SessionConfig):
            dock_widget = self._create_session_config_input(config)
        elif isinstance(config, distance.DetectorConfig):
            dock_widget = self._create_distance_config_input(config)
        elif isinstance(config, presence.DetectorConfig):
            dock_widget = self._create_presence_config_input(config)
        else:
            return

        self._add_dock_widget(dock_widget)
        if animate:
            run_blink_animation(dock_widget)

    def _add_dock_widget(self, dockwidget: QDockWidget) -> None:
        super().addDockWidget(_ALLOWED_DOCK_AREA, dockwidget)

    def _add_dock_widget_below_target(self, target: QDockWidget, new_widget: QDockWidget) -> None:
        super().splitDockWidget(target, new_widget, Qt.Orientation.Vertical)

    def _add_dock_widget_right_of_target(
        self, target: QDockWidget, new_widget: QDockWidget
    ) -> None:
        super().splitDockWidget(target, new_widget, Qt.Orientation.Horizontal)

    def _create_action(
        self,
        icon: QIcon,
        button_label: str,
        on_trigger: t.Callable[[], t.Any],
    ) -> QAction:
        action = QAction(icon, button_label, parent=self)
        action.triggered.connect(on_trigger)
        return action

    def _populate_palette(self, toolbar: QToolBar) -> None:
        toolbar.addAction(
            self._create_action(
                COG(),
                button_label="New Sparse IQ config",
                on_trigger=lambda: self.spawn_input_block(a121.SessionConfig(), animate=False),
            )
        )
        toolbar.addAction(
            self._create_action(
                COG(),
                "New Distance config",
                on_trigger=lambda: self.spawn_input_block(
                    distance.DetectorConfig(), animate=False
                ),
            )
        )
        toolbar.addAction(
            self._create_action(
                COG(),
                "New Presence config",
                on_trigger=lambda: self.spawn_input_block(
                    presence.DetectorConfig(), animate=False
                ),
            )
        )

        toolbar.addSeparator()

        toolbar.addAction(
            self._create_action(
                CHART_BAR(),
                button_label="Power curve",
                on_trigger=lambda: self._add_dock_widget(self._create_power_curve_output()),
            )
        )

        toolbar.addAction(
            self._create_action(
                CHART_LINE(),
                button_label="Power consumption vs. Rate",
                on_trigger=lambda: self._add_dock_widget(
                    self._create_power_consumption_vs_rate_output()
                ),
            )
        )
        toolbar.addAction(
            self._create_action(
                MEMORY(),
                button_label="Memory breakdown",
                on_trigger=lambda: self._add_dock_widget(self._create_memory_breakdown_output()),
            )
        )

        toolbar.addSeparator()

        toolbar.addAction(
            self._create_action(
                INFO(),
                button_label="Info",
                on_trigger=lambda: self._show_info_popup(),
            )
        )

        toolbar.addSeparator()

        toolbar.addWidget(self._compare_cb)

    def _create_session_config_input(self, config: a121.SessionConfig) -> _DockWidget:
        service = session_config_input.SessionConfigInput(self._broker, config)
        return _InputDockWidget(
            service.id_,
            service.fixed_title,
            service.description,
            service,
            service.uninstall_function,
            self._broker,
            parent=self,
        )

    def _create_power_curve_output(self) -> _DockWidget:
        service = power_curve_output.EnergyRegionOutput(self._broker)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )

    def _create_power_consumption_vs_rate_output(self) -> _DockWidget:
        service = power_consumption_vs_rate_output.PowerConsumptionVsRateOutput(self._broker)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )

    def _create_memory_breakdown_output(self) -> _DockWidget:
        service = memory_breakdown_output.MemoryBreakdownOutput(self._broker)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )

    def _create_distance_config_input(self, config: distance.DetectorConfig) -> _DockWidget:
        service = distance_config_input.DistanceConfigInput(self._broker, config)
        return _InputDockWidget(
            service.id_,
            service.fixed_title,
            service.description,
            service,
            service.uninstall_function,
            self._broker,
            parent=self,
        )

    def _create_presence_config_input(self, config: presence.DetectorConfig) -> _DockWidget:
        service = presence_config_input.PresenceConfigInput(self._broker, config)
        return _InputDockWidget(
            service.id_,
            service.fixed_title,
            service.description,
            service,
            service.uninstall_function,
            self._broker,
            parent=self,
        )

    def _show_info_popup(self) -> None:
        txt = "<h2>How to use</h2>"
        txt += _TAB_BRIEF_DESCRIPTION
        txt += "<h2>Disclaimer</h2>"
        txt += _DATA_IS_APPROXIMATE_DESCRIPTION
        QMessageBox.information(self, "Resource Calculator", txt)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import enum
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121, opser
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoConfigBase,
    Controller,
)
from acconeer.exptool.a121.algo.presence._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
)


class DetectionState(enum.Enum):
    """Indicating the current detection state."""

    NO_DETECTION = enum.auto()
    """No detection."""

    DETECTION = enum.auto()
    """Detection."""

    RETENTION = enum.auto()
    """No detection. Retaining detection."""


@attributes_doc
@attrs.mutable(kw_only=True)
class ExampleAppConfig(AlgoConfigBase):
    # Setup parameters
    sensor_to_water_distance: float = attrs.field(default=0.12)
    """Distance from the sensor location to the center of the water jet (m)."""

    water_jet_width: float = attrs.field(default=0.02)
    """Water jet width (m).

    Defines the width of a region, centered around the water jet, where the hand motion is not
    analyzed, in order to exclude the water jet motion from the motion metric.
    """

    measurement_range_end: float = attrs.field(default=0.2)
    """Distance between sensor and end point of measurement range (m)."""

    # Filtering parameters
    filter_time_const: float = attrs.field(default=0.3)
    """Time constant of filter applied during processing.

    A lower value yield a more responsive algorithm. A higher value yield a more robust response.
    """

    threshold: float = attrs.field(default=1.5)
    """Threshold applied to the calculate metric.

    If the metric is above the threshold, a hand movement has been detected.
    """

    detection_retention_duration: float = attrs.field(default=1.0)
    """Detection retention duration (s).

    The detection will be kept as True for this duration, after detection is lost.
    """

    # Sensor config parameters
    hwaas: int = attrs.field(default=32)
    """HWAAS."""

    sweeps_per_frame: int = attrs.field(default=64)
    """Sweeps per frame."""

    sweeps_rate: int = attrs.field(default=1000)
    """Sweep rate."""

    frame_rate: int = attrs.field(default=10)
    """Frame rate."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        return validation_results


opser.register_json_presentable(ExampleAppConfig)


@attrs.frozen(kw_only=True)
class ExtraResult:
    history: npt.NDArray[np.float64] = attrs.field()
    """Detection history."""

    history_time: npt.NDArray[np.float64] = attrs.field()
    """Time vector for detection history."""

    threshold: float = attrs.field()
    """Threshold."""


@attrs.frozen(kw_only=True)
class ExampleAppResult:
    detection_state: DetectionState = attrs.field()
    """The state of the detection."""

    extra_result: ExtraResult = attrs.field()
    """Extra result used for plotting."""


class ExampleApp(Controller[ExampleAppConfig, ExampleAppResult]):
    """Hand motion detection example app.

    :param client: Client
    :param sensor_id: Sensor id
    :param example_app_config: Hand motion detection configuration
    """

    processor: Processor
    HISTORY_LENGTH_S = 10.0

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        example_app_config: ExampleAppConfig,
    ) -> None:
        super().__init__(client=client, config=example_app_config)
        self.sensor_id = sensor_id
        self.config = example_app_config

        self.detection_retention_duration = int(
            round(example_app_config.detection_retention_duration * example_app_config.frame_rate)
        )

        self.started = False

    def start(
        self, recorder: Optional[a121.Recorder] = None, _algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        self._reinitialize_state_variables()

        sensor_config = self._get_sensor_config(self.config)
        self.session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        metadata = self.client.setup_session(self.session_config)
        assert isinstance(metadata, a121.Metadata)

        self.processor = Processor(
            sensor_config=sensor_config,
            processor_config=self._get_processor_config(self.config),
            context=ProcessorContext(estimated_frame_rate=self.config.frame_rate),
            metadata=metadata,
        )

        if _algo_group is None and isinstance(recorder, a121.H5Recorder):
            _algo_group = recorder.require_algo_group("faucet")

        if _algo_group is not None:
            _record_algo_data(_algo_group, self.sensor_id, self.config)

        if recorder is not None:
            self.client.attach_recorder(recorder)

        self.client.start_session()

        self.started = True

    def _reinitialize_state_variables(self) -> None:
        num_points_history = int(self.HISTORY_LENGTH_S * self.config.frame_rate)
        self.history = np.zeros(num_points_history)
        self.history_time = np.linspace(-self.HISTORY_LENGTH_S, 0, num_points_history)
        self.has_detected = False
        self.update_index = 0

    def get_next(self) -> ExampleAppResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processor is not None
        processor_result = self.processor.process(result)

        return self._process_processor_result(processor_result, self.config)

    def _process_processor_result(
        self, result: ProcessorResult, config: ExampleAppConfig
    ) -> ExampleAppResult:
        max_presence_score = max(result.inter_presence_score, result.intra_presence_score)

        # Determine detections state.
        detection_state = DetectionState.NO_DETECTION

        if result.presence_detected:
            self.has_detected = True
            self.update_index_at_detection = self.update_index
            detection_state = DetectionState.DETECTION

        if (
            self.has_detected
            and detection_state is not DetectionState.DETECTION
            and (self.update_index - self.update_index_at_detection)
            < self.detection_retention_duration
        ):
            detection_state = DetectionState.RETENTION

        # Prepare extra result(used for plotting).
        self.history = np.roll(self.history, shift=-1, axis=0)
        self.history[-1] = max_presence_score
        extra_result = ExtraResult(
            history=self.history,
            history_time=self.history_time,
            threshold=config.threshold,
        )

        self.update_index += 1

        return ExampleAppResult(detection_state=detection_state, extra_result=extra_result)

    def stop_detector(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        self.started = False

        return None

    def stop_recorder(self) -> Any:
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        return recorder_result

    def stop(self) -> Any:
        self.stop_detector()
        recorder_result = self.stop_recorder()

        return recorder_result

    @classmethod
    def _get_sensor_config(cls, config: ExampleAppConfig) -> a121.SensorConfig:
        """Convert example app config parameters to sensor config.

        Creates a subsweep configuration with
        - A subsweep between the sensor and the water jet(if it fits).
        - A subsweep between the water jet and the end of the measurement range(if it fits).
        """
        profile = a121.Profile.PROFILE_1
        step_length = 6
        water_jet_half_width = config.water_jet_width / 2.0

        subsweep_configs = []

        # Determine if subsweep between sensor and water jet is feasible
        if ENVELOPE_FWHM_M[profile] < (config.sensor_to_water_distance - water_jet_half_width):
            subsweep_config = a121.SubsweepConfig()
            subsweep_config.start_point = int(
                round(
                    (
                        config.sensor_to_water_distance
                        - ENVELOPE_FWHM_M[profile]
                        - water_jet_half_width
                    )
                    / APPROX_BASE_STEP_LENGTH_M
                )
            )
            subsweep_config.num_points = 1
            subsweep_config.step_length = step_length
            subsweep_config.profile = profile
            subsweep_config.hwaas = config.hwaas
            subsweep_config.receiver_gain = 4
            subsweep_configs.append(subsweep_config)

        # Determine if subsweep after water jet is required
        if (
            ENVELOPE_FWHM_M[profile] + water_jet_half_width + config.sensor_to_water_distance
            < config.measurement_range_end
        ):
            start_m = (
                config.sensor_to_water_distance + water_jet_half_width + ENVELOPE_FWHM_M[profile]
            )
            start_point = int(round(start_m / APPROX_BASE_STEP_LENGTH_M))
            num_points = max(
                1,
                int(
                    round(
                        (config.measurement_range_end - start_m)
                        / (step_length * APPROX_BASE_STEP_LENGTH_M)
                    )
                ),
            )

            subsweep_config = a121.SubsweepConfig()
            subsweep_config.start_point = start_point
            subsweep_config.num_points = num_points
            subsweep_config.step_length = step_length
            subsweep_config.profile = profile
            subsweep_config.hwaas = config.hwaas
            subsweep_config.receiver_gain = 4
            subsweep_configs.append(subsweep_config)

        sensor_config = a121.SensorConfig(subsweeps=subsweep_configs)
        sensor_config.frame_rate = config.frame_rate
        sensor_config.sweep_rate = config.sweeps_rate
        sensor_config.sweeps_per_frame = config.sweeps_per_frame
        sensor_config.inter_frame_idle_state = a121.IdleState.SLEEP
        sensor_config._inter_sweep_idle_state = a121.IdleState.SLEEP

        return sensor_config

    @classmethod
    def _get_processor_config(cls, config: ExampleAppConfig) -> ProcessorConfig:
        return ProcessorConfig(
            intra_enable=True,
            inter_enable=True,
            intra_detection_threshold=config.threshold,
            inter_detection_threshold=config.threshold,
            inter_frame_presence_timeout=None,
            inter_frame_fast_cutoff=3.0,
            inter_frame_slow_cutoff=0.8,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=config.filter_time_const,
            intra_frame_time_const=0.15,
            intra_output_time_const=config.filter_time_const,
        )


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    config: ExampleAppConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    app_config_group = algo_group.create_group("example_app_config")
    opser.serialize(config, app_config_group)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, ExampleAppConfig]:
    sensor_id = int(algo_group["sensor_id"][()])
    config = opser.deserialize(algo_group["example_app_config"], ExampleAppConfig)
    return sensor_id, config



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import abc
from enum import Enum
from typing import Any, Callable, List, Optional

import attrs

from acconeer.exptool.app.new._enums import PluginFamily, PluginGeneration
from acconeer.exptool.app.new.app_model import AppModel, PluginPresetSpec, PluginSpec
from acconeer.exptool.app.new.backend import BackendPlugin, Message

from .plot_plugin_base import PlotPluginBase
from .view_plugin_base import ViewPluginBase


@attrs.frozen(kw_only=True)
class PluginSpecBase(abc.ABC, PluginSpec):
    generation: PluginGeneration = attrs.field()
    key: str = attrs.field()
    title: str = attrs.field()
    docs_link: Optional[str] = attrs.field(default=None)
    description: Optional[str] = attrs.field(default=None)
    family: PluginFamily = attrs.field()
    presets: List[PluginPresetSpec] = attrs.field()
    default_preset_id: Enum = attrs.field()

    @abc.abstractmethod
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin[Any]:
        pass

    @abc.abstractmethod
    def create_view_plugin(self, app_model: AppModel) -> ViewPluginBase:
        pass

    @abc.abstractmethod
    def create_plot_plugin(self, app_model: AppModel) -> PlotPluginBase:
        pass



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import enum
import typing as t


def pretty_dict_line_strs(d: dict[str, t.Any], indent: int = 2, width: int = 24) -> list[str]:
    lines = []
    for k, v in d.items():
        if isinstance(v, enum.Enum):
            v = v.name

        lines.append(f"{'':<{indent}}{k + ' ':.<{width}} {v}")

    return lines



# Copyright (c) Acconeer AB, 2022-2026
# All rights reserved

from __future__ import annotations

import functools
import operator
from typing import Optional

import serial

from acconeer.exptool.flash._device_flasher_base import DeviceFlasherBase
from acconeer.exptool.flash._stm32uart._meta import DEVICE_NAME_TO_ID


STM32_DEFAULT_BAUDRATE = 115200
STM32_UART_TIMEOUT = 2

FLASH_ADDRESS = 0x08000000

DFU_BOOT_DESCRIPTION = {
    "XM125": (
        "<p>"
        "<ol>"
        "<li>Press and hold the <b>DFU</b> button on the board</li>"
        "<li>Press the <b>RESET</b> button (still holding the DFU button)</li>"
        "<li>Release the <b>RESET</b> button</li>"
        "<li>Release the <b>DFU</b> button</li>"
        "</ol>"
        "</p>"
    ),
}

POST_DFU_DESCRIPTION = {}


class Stm32DeviceException(Exception):
    pass


class Stm32FlashException(Exception):
    pass


class Stm32UartFlasher(DeviceFlasherBase):
    def flash(self, serial_device, device_name, image_path, progress_callback=None):
        if device_name not in DEVICE_NAME_TO_ID:
            msg = f"Unknown device '{device_name}'"
            raise Stm32DeviceException(msg)

        device_chip_id = DEVICE_NAME_TO_ID[device_name]

        flasher = Stm32FlashProtocol(serial_device.port)
        flasher.sync()
        chip_id = flasher.get_id()
        if chip_id != device_chip_id:
            msg = f"Incorrect STM32 chip_id 0x{chip_id:x}"
            raise Stm32DeviceException(msg)

        with open(image_path, "rb") as image_file:
            image_data = image_file.read()
            flasher.extended_erase()
            flasher.write_image(image_data, FLASH_ADDRESS, progress_callback=progress_callback)
            flasher.write_unprotect()

        flasher.close()

    @staticmethod
    def get_boot_description(device_name: str) -> Optional[str]:
        description = (
            "Please visit https://developer.acconeer.com to find the latest flash "
            f"instructions for {device_name}."
        )

        if device_name in DFU_BOOT_DESCRIPTION:
            description = (
                f"<p>To flash the {device_name} it needs to be put in bootloader mode:</p>"
                f"{DFU_BOOT_DESCRIPTION[device_name]}"
            )

        return description

    @staticmethod
    def get_post_dfu_description(device_name: str) -> Optional[str]:
        description = None
        if device_name in POST_DFU_DESCRIPTION:
            description = (
                f"<p>The {device_name} needs to be reset before it can be used:</p>"
                f"{POST_DFU_DESCRIPTION[device_name]}"
            )

        return description


class Stm32FlashProtocol:
    ACK = 0x79
    NACK = 0x1F

    def __init__(self, port):
        self._port = port
        self._uart = UartComm(self._port)

    def close(self):
        if self._uart is not None:
            self._uart.close()
            self._uart = None

    def _wait_for_ack_or_nack(self):
        reply_array = self._uart.read(length=1)
        if len(reply_array) > 0:
            reply = reply_array[0]
            if reply in [self.ACK, self.NACK]:
                return reply
            else:
                msg = f"Unknown reply from STM32 bootloader [0x{reply:x}]"
                raise Stm32FlashException(msg)
        msg = "Device did not reply any data"
        raise Stm32FlashException(msg)

    def _wait_for_ack(self):
        reply = self._wait_for_ack_or_nack()
        if reply == self.NACK:
            msg = "Device replied NACK to command"
            raise Stm32FlashException(msg)

    def _command(self, cmd_byte):
        cmd_xor_byte = cmd_byte ^ 0xFF
        cmd = bytearray([cmd_byte, cmd_xor_byte])
        self._uart.write(cmd)
        self._wait_for_ack()

    def _read_bytes(self, length):
        byte_array = self._uart.read(length=length)
        if len(byte_array) != length:
            Stm32FlashException("Device did not reply correct amount of data")
        return byte_array

    def _checksum(self, array):
        return functools.reduce(operator.xor, array, 0)

    def _send_address(self, address):
        data_bytes = bytearray(address.to_bytes(length=4, byteorder="big"))
        checksum = self._checksum(data_bytes)
        data_bytes.append(checksum)
        self._uart.write(data_bytes)
        self._wait_for_ack()

    def _send_data(self, data_bytes):
        tx_data = bytearray([len(data_bytes) - 1])
        tx_data.extend(data_bytes)
        checksum = self._checksum(tx_data)
        tx_data.append(checksum)
        self._uart.write(tx_data)
        self._wait_for_ack()

    def _write_memory(self, address, byte_array):
        WRITE_MEMORY = 0x31

        if len(byte_array) > 256:
            Stm32FlashException("Memory chunk to large")

        while (len(byte_array) % 4) != 0:
            byte_array.extend(0xFF)

        self._command(WRITE_MEMORY)
        self._send_address(address)
        self._send_data(byte_array)

    def sync(self):
        CMD_SYNC = bytearray([0x7F])
        # Try so SYNC twice before failing
        for _ in range(0, 2):
            try:
                self._uart.write(CMD_SYNC)
                self._wait_for_ack_or_nack()
                return
            except Stm32FlashException:
                pass
        msg = f"DFU synchronization failed for '{self._port}', make sure device is in DFU mode"
        raise Stm32FlashException(msg)

    def get_id(self):
        GET_ID = 0x02
        self._command(GET_ID)
        len_bytes = self._read_bytes(1)
        len = int.from_bytes(len_bytes, byteorder="big")
        id_bytes = self._read_bytes(len + 1)
        id = int.from_bytes(id_bytes, byteorder="big")
        self._wait_for_ack()
        return id

    def extended_erase(self):
        EXTENDED_ERASE = 0x44
        self._command(EXTENDED_ERASE)
        self._uart.write([0xFF, 0xFF, 0x00])
        self._wait_for_ack()

    def write_image(self, byte_array, flash_address, progress_callback=None):
        MAX_CHUNK_SIZE = 256

        for pos in range(0, len(byte_array), MAX_CHUNK_SIZE):
            if progress_callback is not None:
                progress_callback(int(100 * pos / len(byte_array)))

            chunk_len = min(MAX_CHUNK_SIZE, len(byte_array) - pos)
            chunk_array = bytearray(byte_array[pos : pos + chunk_len])

            self._write_memory(flash_address + pos, chunk_array)

        if progress_callback is not None:
            progress_callback(100, True)

    def write_unprotect(self):
        WRITE_UNPROTECT = 0x73
        self._command(WRITE_UNPROTECT)
        self._wait_for_ack()


class UartComm:
    def __init__(self, port):
        self._ser = serial.Serial(
            port=port,
            baudrate=STM32_DEFAULT_BAUDRATE,
            parity="E",
            rtscts=False,
            timeout=STM32_UART_TIMEOUT,
        )
        self._ser.reset_input_buffer()

    def close(self):
        if self._ser is not None:
            self._ser.close()
            self._ser = None

    def write(self, data):
        self._ser.write(data)

    def read(self, length=1):
        return bytearray(self._ser.read(length))



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved
from __future__ import annotations

import typing as t

import attrs
import typing_extensions as te

from .message import Message, ParseError


LogLevelStr = t.Union[
    te.Literal["ERROR"],
    te.Literal["WARNING"],
    te.Literal["INFO"],
    te.Literal["VERBOSE"],
    te.Literal["DEBUG"],
]


class LogMessageHeader(te.TypedDict):
    level: LogLevelStr
    timestamp: int
    module: str
    log: str


@attrs.frozen
class LogMessage(Message):
    message: ServerLog

    @classmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> LogMessage:
        t.cast(LogMessageHeader, header)

        if header["status"] == "log":
            return cls(
                ServerLog(
                    level=header["level"],
                    timestamp=header["timestamp"],
                    module=header["module"],
                    log=header["log"],
                )
            )
        raise ParseError


@attrs.frozen(kw_only=True)
class ServerLog:
    level: str
    timestamp: int
    module: str
    log: str



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from .buffered_link import BufferedLink


class NullLinkError(RuntimeError):
    pass


class NullLink(BufferedLink):
    ERROR = NullLinkError("Link is undetermined.")
    """Link null object.

    :raises: ``RuntimeError`` if any of its methods is called.
    """

    def connect(self) -> None:
        raise self.ERROR

    @property
    def timeout(self) -> float:
        raise self.ERROR

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        raise self.ERROR

    def _update_timeout(self) -> None:
        pass

    def recv(self, num_bytes: int) -> bytes:
        raise self.ERROR

    def send(self, bytes_: bytes) -> None:
        raise self.ERROR

    def disconnect(self) -> None:
        raise self.ERROR

    def recv_until(self, byte_sequence: bytes) -> bytes:
        raise self.ERROR



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import typing_extensions as te

from PySide6.QtWidgets import QWidget

from acconeer.exptool.app.new._enums import ConnectionState, PluginGeneration, PluginState
from acconeer.exptool.app.new.app_model import AppModel


class PolicyFunction(te.Protocol):
    def __call__(self, app_model: AppModel, *, extra_condition: bool = True) -> bool: ...


def start_button_enabled(app_model: AppModel, *, extra_condition: bool = True) -> bool:
    return (
        app_model.plugin_state == PluginState.LOADED_IDLE
        and app_model.connection_state == ConnectionState.CONNECTED
        and app_model.plugin_generation == PluginGeneration.A121
        and bool(app_model.connected_sensors)
        and app_model.backend_plugin_state is not None
        and getattr(app_model.backend_plugin_state, "ready", True)
        and extra_condition
    )


def stop_button_enabled(app_model: AppModel, *, extra_condition: bool = True) -> bool:
    return app_model.plugin_state == PluginState.LOADED_BUSY and extra_condition


def config_editor_enabled(app_model: AppModel, *, extra_condition: bool = True) -> bool:
    return app_model.plugin_state == PluginState.LOADED_IDLE and extra_condition


def apply_enabled_policy(
    policy_func: PolicyFunction,
    app_model: AppModel,
    widgets: list[QWidget],
    *,
    extra_condition: bool = True,
) -> None:
    for widget in widgets:
        widget.setEnabled(policy_func(app_model, extra_condition=extra_condition))



#######################################
# Copyright (c) Acconeer AB, 2025-2026
# All rights reserved
#######################################

"""DFU config for the XC120 board firmware"""

from ._bootloader_tool import BootloaderTool


XC120_VID = 0x0483
XC120_EXPLORATION_SERVER_WINUSB_PID = 0xA449
XC120_BOOTLOADER_PID = 0xA41D
XC120_PROTOCOL_SERVER_PID = 0xA42C
XC120_PROTOCOL_SERVER_ID = 0xF2


class BootLoaderXC120(BootloaderTool):
    def __init__(self):
        super().__init__(device_vid=XC120_VID, bootloader_pid=XC120_BOOTLOADER_PID)

    def device_enter_dfu(self, device_pid, device_port):
        if device_pid == XC120_BOOTLOADER_PID:
            success = True
        elif device_pid == XC120_EXPLORATION_SERVER_WINUSB_PID:
            success = self.exploration_server_to_dfu(
                XC120_VID, XC120_EXPLORATION_SERVER_WINUSB_PID
            )
        elif device_pid == XC120_PROTOCOL_SERVER_PID:
            success = self.protocol_server_to_dfu(device_port, XC120_PROTOCOL_SERVER_ID)
        else:
            success = False
        return success



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved
from __future__ import annotations

import typing as t
from enum import Enum

import attrs
import numpy as np

from . import core
from .registry_persistor import RegistryPersistor


_T = t.TypeVar("_T")
_S = t.TypeVar("_S")


@RegistryPersistor.register_persistor
class StringPersistor(core.Persistor):
    """
    Persists a string by encoding/decoding to/from bytes
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return __type is str

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, str):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", str)

        self.create_own_dataset(data=data.encode())

    def _load(self) -> str:
        encoded_string = self.dataset[()]

        self.assert_not_empty(encoded_string)

        return bytes.decode(encoded_string)


@RegistryPersistor.register_persistor
class EnumPersistor(core.Persistor):
    """
    Persists an Enum value by saving/loading its encoded name to a dataset
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return core.is_subclass(__type, Enum)

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, Enum):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", Enum)

        self.create_own_dataset(data=data.name.encode())

    def _load(self) -> Enum:
        enum_name_in_bytes = self.dataset[()]

        self.assert_not_empty(enum_name_in_bytes)

        enum_type = self.type_tree.data

        return enum_type[bytes.decode(enum_name_in_bytes)]  # type: ignore[index, no-any-return]


@RegistryPersistor.register_persistor
class NonePersistor(core.Persistor):
    """
    Persists a None by saving/"loading" an empty dataset

    See: https://docs.h5py.org/en/stable/high/dataset.html#dataset-empty
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return core.is_subclass(__type, type(None))

    def _save(self, data: t.Any) -> None:
        if data is not None:
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", type(None))

        self.create_own_dataset(data, dtype="f")

    def _load(self) -> None:
        should_be_empty = self.dataset[()]

        self.assert_empty(should_be_empty)


@RegistryPersistor.register_persistor
class FloatPersistor(core.Persistor):
    """
    Persists floats.

    Also persists ints, but it is transformed to a float before saving
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return __type is float

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, (float, int)):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", float, int)

        if np.shape(data) != ():
            raise core.TypeMissmatchError

        self.create_own_dataset(data=float(data))

    def _load(self) -> float:
        should_be_float = self.dataset[()]

        self.assert_not_empty(should_be_float)

        return float(should_be_float)


@RegistryPersistor.register_persistor
class BoolPersistor(core.Persistor):
    """
    Persists bools
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return __type is bool

    def _save(self, data: t.Any) -> None:
        if data not in {True, False}:
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", bool, np.bool_)

        if np.shape(data) != ():
            raise core.TypeMissmatchError

        self.create_own_dataset(data=data)

    def _load(self) -> bool:
        should_be_bool = self.dataset[()]

        self.assert_not_empty(should_be_bool)

        return bool(should_be_bool)


@RegistryPersistor.register_persistor
class IntPersistor(core.Persistor):
    """
    Persists ints
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return __type is int

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, (int, np.integer)):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", int, np.integer)

        if np.shape(data) != ():
            raise core.TypeMissmatchError

        self.create_own_dataset(data=data)

    def _load(self) -> int:
        should_be_int = self.dataset[()]

        self.assert_not_empty(should_be_int)

        return int(should_be_int)


@RegistryPersistor.register_persistor
class ListPersistor(core.Persistor):
    """
    Persists lists of arbitrary type.

    Loading/saving recurses on each element in the list / member of the group
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_list, *_ = core.unwrap_generic(__type)
        return should_be_list is list

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, list):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", list)

        (element_type_tree,) = self.type_tree.children.values()

        element_group = self.require_own_group()

        for i, element in enumerate(data):
            RegistryPersistor(element_group, str(i), element_type_tree).save(element)

    def _load(self) -> t.List[_T]:
        (element_type_tree,) = self.type_tree.children.values()

        return [
            RegistryPersistor(self.group, key, element_type_tree).load()
            for key in sorted(self.group.keys(), key=int)
        ]


@RegistryPersistor.register_persistor
class DictPersistor(core.Persistor):
    """
    Persists dicts

    dicts gets persisted as entries with key and value members. This allows
    both keys and values to be of arbitrary type
    """

    VALUE_GROUP_KEY = "val"
    KEY_GROUP_KEY = "key"

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_dict, *_ = core.unwrap_generic(__type)
        return should_be_dict is dict

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, dict):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", dict)

        (key_type_tree, value_type_tree) = self.type_tree.children.values()

        dict_group = self.require_own_group()

        for i, (key, value) in enumerate(data.items()):
            entry_group = dict_group.create_group(str(i))
            RegistryPersistor(entry_group, self.VALUE_GROUP_KEY, value_type_tree).save(value)
            RegistryPersistor(entry_group, self.KEY_GROUP_KEY, key_type_tree).save(key)

    def _load(self) -> t.Dict[_S, _T]:
        (key_type_tree, value_type_tree) = self.type_tree.children.values()

        return {
            RegistryPersistor(self.group[entry_index], self.KEY_GROUP_KEY, key_type_tree).load(): (
                RegistryPersistor(
                    self.group[entry_index], self.VALUE_GROUP_KEY, value_type_tree
                ).load()
            )
            for entry_index in sorted(self.group.keys(), key=int)
        }


@RegistryPersistor.register_persistor
class TuplePersistor(core.Persistor):
    """
    Persists non-variadic tuples with arbitrary types

    I.e. Tuple[str, int, SomeClass] is persisted like expected
    but Tulpe[int, ...] is not supported
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_tuple, *_ = core.unwrap_generic(__type)
        return should_be_tuple is tuple

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, tuple):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", tuple)

        element_group = self.require_own_group()

        for i, (element_subtree, element) in enumerate(
            zip(self.type_tree.children.values(), data)
        ):
            RegistryPersistor(element_group, str(i), element_subtree).save(element)

    def _load(self) -> t.Tuple[t.Any, ...]:
        return tuple(
            RegistryPersistor(self.group, key, element_type_tree).load()
            for key, element_type_tree in zip(self.group.keys(), self.type_tree.children.values())
        )


@RegistryPersistor.register_persistor
class UnionPersistor(core.Persistor):
    """
    Persists union types

    Each member of the Union gets an attempt to persist before exiting
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_union, *_ = core.unwrap_generic(__type)
        return should_be_union is t.Union

    def _save(self, data: t.Any) -> None:
        errors = []
        for member_type_tree in self.type_tree.children.values():
            try:
                RegistryPersistor(self.parent_group, self.name, member_type_tree).save(data)
            except core.SaveError as save_error:
                errors += [save_error]
                continue
            else:
                return

        msg = f"Could not save instance {data!r:.100}... of type {type(data)}"
        raise core.SaveErrorGroup(msg, errors)

    def _load(self) -> t.Any:
        errors = []
        for name, subtree in self.type_tree.children.items():
            try:
                res = RegistryPersistor(self.parent_group, self.name, subtree).load()
            except core.LoadError as load_error:
                errors += [load_error]
                continue
            else:
                return res

        child_types = tuple(st.data for st in self.type_tree.children.values())
        msg = f"Could not load any of {child_types} from {self.group}"
        raise core.LoadErrorGroup(msg, errors)


@RegistryPersistor.register_persistor
class AttrsInstancePersistor(core.Persistor):
    """
    Persists attrs instances
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return attrs.has(__type)

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, self.type_tree.data):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", self.type_tree.data)

        instance_group = self.require_own_group()
        for member_name, member_type_tree in self.type_tree.children.items():
            RegistryPersistor(instance_group, member_name, member_type_tree).save(
                getattr(data, member_name)
            )

    def _load(self) -> attrs.AttrsInstance:
        attrs_class = self.type_tree.data

        assert attrs.has(attrs_class)

        return attrs_class(
            **{
                member_name.strip("_"): RegistryPersistor(
                    self.group, member_name, member_type_tree
                ).load()
                for member_name, member_type_tree in self.type_tree.children.items()
            }
        )


@RegistryPersistor.register_persistor
class NumpyPersistor(core.Persistor):
    """
    Persists numpy arrays/scalars
    """

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        return core.is_ndarray(__type) or core.is_subclass(__type, np.generic)

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, (np.ndarray, np.generic)):
            raise core.TypeMissmatchError.wrong_type_encountered(data, "?", np.ndarray)

        self.create_own_dataset(data=data)

    def _load(self) -> t.Any:
        should_be_np = self.dataset[()]

        self.assert_not_empty(should_be_np)

        return should_be_np



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
"""
This module defines the primitives of an event system.
The event system consists of 2 distinct types of actors;
the EventBroker and Services.

The EventBroker provides a publisher/subscriber implementation
for Services.

A consumer-Service subscribes to topics of interest
(in the variable INTERESTS) and then declares its interests to the
EventBroker by installing itself (EventBroker.install_service).

A producer-Service offers events of arbitrary type to the
EventBroker which will in turn dispatch that event to any
interested consumer Services.

Producer Services may want to put an unique identifiers
on its events so that consumers are able to tell events from
different producers apart (this allows duplicated producer services).
This is done via 'EventBroker.install_identified_service'.
"""

from __future__ import annotations

import collections
import contextlib
import itertools
import logging
import typing as t
from copy import copy

import attrs
import typing_extensions as te


log = logging.getLogger(__name__)


class Service(te.Protocol):
    INTERESTS: t.ClassVar[set[type]]
    """Defines what events this service is interested in"""

    description: t.ClassVar[str]
    """A description that explains the block"""

    window_title: str
    """The window title"""

    uninstall_function: t.Callable[[], None]
    """Uninstalls this service from the EventBroker (returned when installing)"""

    def handle_event(self, event: t.Any) -> None:
        """
        Handles an incoming event.
        'event' can be of any type defined in INTERESTS.
        """
        ...


class _IdLedger:
    def __init__(self) -> None:
        self._taken_ids: set[str] = set()

    def borrow_numbered_id(self, prefix: str) -> str:
        for serial_number in itertools.count(1):
            candidate = f"{prefix}-{serial_number}"
            if candidate not in self._taken_ids:
                self._taken_ids.add(candidate)
                return candidate

        # since itertools.count returns an infinite iterator,
        # we should never end up here.
        raise RuntimeError

    def borrow_id(self, id_: str) -> str:
        if id_ in self._taken_ids:
            return self.borrow_numbered_id(id_)
        else:
            self._taken_ids.add(id_)
            return id_

    def return_id(self, id_: str) -> None:
        self._taken_ids.remove(id_)


@attrs.frozen
class IdentifiedServiceUninstalledEvent:
    """
    This event is offered when an identified service is removed,
    allowing consumers to clean up resources related to that
    identified service.
    """

    id_: str


@attrs.frozen
class ChangeIdEvent:
    old_id: str
    new_id: str


class EventBroker:
    def __init__(self, event_log_capacity: int = 100) -> None:
        self._topic_callbacks: dict[type, list[t.Callable[[t.Any], None]]] = {}
        self._event_log: collections.deque[t.Any] = collections.deque(maxlen=event_log_capacity)
        self._id_pool = _IdLedger()

    def install_service(self, service: Service) -> None:
        """
        Installs a service, subscribing its handle_event to
        all topics in INTERESTS

        returns a function that uninstalls the passed service
        """
        for interest in service.INTERESTS:
            self._topic_callbacks.setdefault(interest, []).append(service.handle_event)

    def uninstall_service(self, service: Service) -> None:
        log.debug(f"Uninstalling service of type {type(service)}")

        for interest in service.INTERESTS:
            with contextlib.suppress(KeyError, ValueError):
                self._topic_callbacks[interest].remove(service.handle_event)

    def change_id(self, new_id: str, old_id: str) -> str:
        new_id = self._id_pool.borrow_id(new_id)

        self._id_pool.return_id(old_id)

        return new_id

    def install_identified_service(self, service: Service, prefix: str) -> str:
        """
        Installs a service that requires an id, subscribing its handle_event to
        all topics in INTERESTS

        returns a function that uninstalls the passed service and the id
        """
        service_id = self._id_pool.borrow_numbered_id(prefix)

        for interest in service.INTERESTS:
            self._topic_callbacks.setdefault(interest, []).append(service.handle_event)

        return service_id

    def uninstall_identified_service(self, service: Service, service_id: str) -> None:
        self.uninstall_service(service)

        self._id_pool.return_id(service_id)

        self.offer_event(IdentifiedServiceUninstalledEvent(service_id))

    def brief_service(self, service: Service) -> None:
        """
        Briefs the service by calling its handle_event with
        the most recent interesting events.
        """
        for event in self._event_log:
            if type(event) in service.INTERESTS:
                service.handle_event(event)

    def offer_event(self, event: t.Any) -> None:
        """
        Offer an event to interested services
        """
        log.debug(f"event with type {type(event)} was offered")
        subscribers_of_event = copy(self._topic_callbacks.get(type(event), []))

        self._event_log.append(event)

        if subscribers_of_event:
            for subscriber in subscribers_of_event:
                subscriber(event)
        else:
            log.debug(f"No subscribers of event with type {type(event)}")



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from .client import Client, ClientCreationError, ClientError
from .communication_protocol import CommunicationProtocol, Message, ParseError
from .links import (
    BufferedLink,
    ExploreSerialLink,
    NullLink,
    NullLinkError,
    SerialLink,
    SerialProcessLink,
    SocketLink,
    USBLink,
)
from .message_stream import MessageStream



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

# THIS FILE IS AUTOMATICALLY GENERATED - DO NOT EDIT
from __future__ import annotations

import logging
import queue
import time
from typing import List


_LOG = logging.getLogger(__name__)


class CommandFailed(Exception):
    pass


class Packet:
    """
    Base class for all packages transmitted over UART protocol.
    """

    PROTOCOL_START_MARKER = 0xCC
    PROTOCOL_END_MARKER = 0xCD

    def __init__(self, payload):
        self.payload = payload

    @property
    def packet_type(self):
        raise NotImplementedError

    def get_byte_array(self):
        """
        Packet the data for this packet into a bytearray()
        that can be transmitted over UART.
        """
        data = bytearray([])
        data.append(Packet.PROTOCOL_START_MARKER)
        packet_length = int(len(self.payload))
        data.extend(packet_length.to_bytes(2, byteorder="little"))
        data.append(self.packet_type)
        data.extend(self.payload)
        data.append(Packet.PROTOCOL_END_MARKER)
        return data


class BlCommandResponsePacket(Packet):
    packet_type = 0xF3

    def __init__(self, payload):
        self.command_id = int.from_bytes(payload[0:2], byteorder="little")
        self.command_payload = payload[2:]
        super().__init__(payload)

    def get_command_packet(self):
        return _command_packets.get(self.command_id)(self.command_payload)


class BlCommandRequestPacket(Packet):
    packet_type = 0xF3

    def __init__(self, command_payload):
        command_payload[0:0] = self.command_id.to_bytes(2, byteorder="little")
        super().__init__(command_payload)


class GetLastErrorRequestPacket(BlCommandRequestPacket):
    command_id = 0x0002

    def __init__(
        self,
    ):
        payload = bytearray()
        super().__init__(payload)


class GetLastErrorResponsePacket:
    command_id = 0x0002

    def __init__(self, payload):
        self.payload = payload

    def get_status(self):
        return self.payload[0]

    def get_response_data(self):
        return self.payload[1:].decode("ascii")


class GetAppSwVersionRequestPacket(BlCommandRequestPacket):
    command_id = 0x0101

    def __init__(
        self,
    ):
        payload = bytearray()
        super().__init__(payload)


class GetAppSwVersionResponsePacket:
    command_id = 0x0101

    def __init__(self, payload):
        self.payload = payload

    def get_status(self):
        return self.payload[0]

    def get_version(self):
        return self.payload[1:].decode("ascii")


class GetAppSwNameRequestPacket(BlCommandRequestPacket):
    command_id = 0x0102

    def __init__(
        self,
    ):
        payload = bytearray()
        super().__init__(payload)


class GetAppSwNameResponsePacket:
    command_id = 0x0102

    def __init__(self, payload):
        self.payload = payload

    def get_status(self):
        return self.payload[0]

    def get_name(self):
        return self.payload[1:].decode("ascii")


class IsImageErasedRequestPacket(BlCommandRequestPacket):
    command_id = 0xF000

    def __init__(self, image_size):
        payload = bytearray()
        payload.extend(image_size.to_bytes(4, byteorder="little", signed=True))
        super().__init__(payload)


class IsImageErasedResponsePacket:
    command_id = 0xF000

    def __init__(self, payload):
        self.payload = payload
        assert len(payload) == 2

    def get_status(self):
        return self.payload[0]

    def get_is_erased(self):
        return self.payload[1] != 0


class ImageEraseRequestPacket(BlCommandRequestPacket):
    command_id = 0xF001

    def __init__(self, image_size):
        payload = bytearray()
        payload.extend(image_size.to_bytes(4, byteorder="little", signed=True))
        super().__init__(payload)


class ImageEraseResponsePacket:
    command_id = 0xF001

    def __init__(self, payload):
        self.payload = payload
        assert len(payload) == 1

    def get_status(self):
        return self.payload[0]


class ImageWriteBlockRequestPacket(BlCommandRequestPacket):
    command_id = 0xF002

    def __init__(self, offset, data):
        payload = bytearray()
        payload.extend(offset.to_bytes(4, byteorder="little", signed=True))
        payload.extend(data)
        super().__init__(payload)


class ImageWriteBlockResponsePacket:
    command_id = 0xF002

    def __init__(self, payload):
        self.payload = payload
        assert len(payload) == 1

    def get_status(self):
        return self.payload[0]


class ResetRequestPacket(BlCommandRequestPacket):
    command_id = 0xFFFE

    def __init__(
        self,
    ):
        payload = bytearray()
        super().__init__(payload)


_command_packets = {
    0x0002: GetLastErrorResponsePacket,
    0x0101: GetAppSwVersionResponsePacket,
    0x0102: GetAppSwNameResponsePacket,
    0xF000: IsImageErasedResponsePacket,
    0xF001: ImageEraseResponsePacket,
    0xF002: ImageWriteBlockResponsePacket,
}


class BLCommunication:
    PROTOCOL_OVERHEAD = 5
    PACKET_TIMEOUT = 0.2
    PROTOCOL_TIMEOUT = 2

    def __init__(self, serial_device):
        self._ser = serial_device
        self._ser.timeout = self.PACKET_TIMEOUT

    def _send_packet(self, packet):
        data = packet.get_byte_array()
        _LOG.debug("Sending packet %s", data)
        self._ser.write(data)

    def _read_response(self):
        response_bytes = bytearray()
        length = 0
        start = time.monotonic()
        while len(response_bytes) < (length + self.PROTOCOL_OVERHEAD):
            data = self._ser.read()
            response_bytes.extend(data)
            if len(response_bytes) > 0 and response_bytes[0] != Packet.PROTOCOL_START_MARKER:
                msg = f"Bad response, start marker '0x{response_bytes[-1]:x}'"
                raise CommandFailed(msg)

            if length == 0 and len(response_bytes) > 2:
                length = response_bytes[1] | response_bytes[2] << 8

            if (time.monotonic() - start) > self.PROTOCOL_TIMEOUT:
                msg = "Timeout when receiving data from device"
                raise CommandFailed(msg)

        if response_bytes[-1] != Packet.PROTOCOL_END_MARKER:
            msg = f"Bad response, end marker '0x{response_bytes[-1]:x}'"
            raise CommandFailed(msg)

        if response_bytes[3] != BlCommandResponsePacket.packet_type:
            msg = f"Bad response, packet type '0x{response_bytes[3]:x}'"
            raise CommandFailed(msg)

        return BlCommandResponsePacket(response_bytes[4:-1])

    def _execute_request(self, packet):
        try:
            self._send_packet(packet)
            response = self._read_response()
            return response.get_command_packet()
        except queue.Empty:
            _LOG.error("read timeout")
            msg = "_execute_request timeout"
            raise TimeoutError(msg)

    def get_last_error(self):
        response = self._execute_request(GetLastErrorRequestPacket())
        status = response.get_status()
        response_data = response.get_response_data()
        if status != 0:
            error_msg = self.get_last_error()
            msg = f"Command failed with code {status} [{error_msg}]"
            raise CommandFailed(msg)
        return response_data

    def get_app_sw_version(self):
        response = self._execute_request(GetAppSwVersionRequestPacket())
        status = response.get_status()
        version = response.get_version()
        if status != 0:
            error_msg = self.get_last_error()
            msg = f"Command failed with code {status} [{error_msg}]"
            raise CommandFailed(msg)
        return version

    def get_app_sw_name(self):
        response = self._execute_request(GetAppSwNameRequestPacket())
        status = response.get_status()
        name = response.get_name()
        if status != 0:
            error_msg = self.get_last_error()
            msg = f"Command failed with code {status} [{error_msg}]"
            raise CommandFailed(msg)
        return name

    def is_image_erased(self, image_size: int):
        response = self._execute_request(IsImageErasedRequestPacket(image_size))
        status = response.get_status()
        is_erased = response.get_is_erased()
        if status != 0:
            error_msg = self.get_last_error()
            msg = f"Command failed with code {status} [{error_msg}]"
            raise CommandFailed(msg)
        return is_erased

    def image_erase(self, image_size: int):
        response = self._execute_request(ImageEraseRequestPacket(image_size))
        status = response.get_status()
        if status != 0:
            error_msg = self.get_last_error()
            msg = f"Command failed with code {status} [{error_msg}]"
            raise CommandFailed(msg)

    def image_write_block(self, offset: int, data: List[int]):
        response = self._execute_request(ImageWriteBlockRequestPacket(offset, data))
        status = response.get_status()
        if status != 0:
            error_msg = self.get_last_error()
            msg = f"Command failed with code {status} [{error_msg}]"
            raise CommandFailed(msg)

    def reset(self):
        self._send_packet(ResetRequestPacket())
        try:
            self._ser.read()  # Dummy read to make sure the previous write is finished
        except Exception:  # noqa: S110
            pass
